#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import CallPut, AverageMethod, AsianAveSubstitution
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base import BiTreeEngine
from pricelib.common.time import global_evaluation_date


class BiTreeAsianEngine(BiTreeEngine):
    """算术平均结算价(收盘价)亚式二叉树定价引擎
    不支持几何平均，不支持增强亚式，不支持平均执行价"""

    def __init__(self, stoch_process: StochProcessBase = None, tree_branches=100, n_samples=200, *,
                 s=None, r=None, q=None, vol=None):
        """初始化算术平均亚式二叉树定价引擎
        Args:
            stoch_process: StochProcessBase，随机过程
            tree_branches: int，二叉树每年的分支数
            n_samples: int，在每个树节点历史路径平均股价的最大值和最小值之间，插入数值点的数量
       在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, tree_branches, s=s, r=r, q=q, vol=vol)
        self.n_samples = n_samples
        # 以下为计算过程的中间变量
        self.prod = None  # Product产品对象
        self.n_step = None  # 二叉树的总分支数
        self.limited_price = None  # 兼容增强亚式的限制价格变量

    def cal_path(self, s_path, obs_start_step, obs_end_step, n_samples):
        """计算每个树节点对应的历史路径平均值的极大值s_max_path与极小值s_min_path，在其中插入固定数量n_samples的数值点，
                        最后返回一个三维数组s_path，其第一维数是n_samples，第二和第三维数等同于树s的第一维和第二维(tree_branches)。
        原理:
        假设第j层树节点分支，(根节点j=0)，每层第i个节点(该层最低股价处i=0)
        不妨设期初(j=0)就开始观察平均收盘价，则到第j层的时候，历史价格路径有(j+1)个节点
        那么历史路径平均值的极大值，前(i+1)层时的价格都一直处于当时股价的最顶端，之后开始逐点下降，直到到达第j层时，下降到第i个节点
                       也就是该价格路径由(i+1)个[1, u, u^2, ..., u^i]和(j-i)个 u^i * d * [1, d, d^2, ...,d^(j-i-1)]构成
        同理历史路径平均值的极小值，前(j-i)层时的价格都一直处于当时股价的最底端，之后开始逐点上升，直到到达第j层时，上升到第i个节点
                       也就是该价格路径由(j-i+1)个[1, d, d^2, ..., d^(j-i)]和i个 d^(j-i) * u * [1, u, u^2, ...,u^(i-1)]构成
        本方法就是按照这种规律进行累加，最后除以每层对应的价格总数(j+1)，即可得到历史路径平均值的极大极小值
        Args:
            s_path: np.ndarray，上三角结构的标的资产价格CRR二叉树，二维数组，标的价格x时间分支
            obs_start_step: float，观察平均的开始时间点，对应的二叉树分支索引
            obs_end_step: float，观察平均的终止时间点，对应的二叉树分支索引
            n_samples: int，在每个树节点历史路径平均股价的最大值和最小值之间，插入数值点的数量
        Returns:
            s_path: np.ndarray，三维数组，其第一维数是n_samples，第二和第三维数等同于树s的第一维和第二维
                    相当于n_samples个二叉树，分别对应了历史路径平均值的极大值s_max_path与极小值s_min_path中间的一个数值点的二叉树
        """
        s_max_path = np.array(s_path)
        s_min_path = np.array(s_path)
        for j in range(self.n_step + 1):
            if j < obs_start_step:
                continue
            if j == obs_start_step:  # 观察起点，开始记录历史价格路径之和的最小值和历史价格路径之和的的最大值
                if self.prod.callput == CallPut.Call:
                    if self.prod.ave_method == AverageMethod.Arithmetic:
                        s_max_path[:j + 1, j] = np.maximum(s_path[:j + 1, j], self.limited_price)
                        s_min_path[:j + 1, j] = np.maximum(s_path[:j + 1, j], self.limited_price)
                    elif self.prod.ave_method == AverageMethod.Geometric:
                        s_max_path[:j + 1, j] = np.maximum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                        s_min_path[:j + 1, j] = np.maximum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                elif self.prod.callput == CallPut.Put:
                    if self.prod.ave_method == AverageMethod.Arithmetic:
                        s_max_path[:j + 1, j] = np.minimum(s_path[:j + 1, j], self.limited_price)
                        s_min_path[:j + 1, j] = np.minimum(s_path[:j + 1, j], self.limited_price)
                    elif self.prod.ave_method == AverageMethod.Geometric:
                        s_max_path[:j + 1, j] = np.minimum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                        s_min_path[:j + 1, j] = np.minimum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                continue
            if j > obs_end_step:  # 观察终点，结束观察平均价格，后续节点的历史价格路径之和的最小值和历史价格路径之和的的最大值都等于上一天顺延
                s_max_path[0, j] = s_max_path[0, j - 1]
                s_max_path[1:j + 1, j] = s_max_path[:j, j - 1]
                s_min_path[j, j] = s_min_path[j - 1, j - 1]
                s_min_path[:j, j] = s_min_path[:j, j - 1]
            else:  # 观察起始点之间，记录历史价格路径之和的最小值和历史价格路径之和的的最大值
                s_max_path[0, j] = s_max_path[0, j - 1]
                s_max_path[1:j + 1, j] = s_max_path[:j, j - 1]
                s_min_path[j, j] = s_min_path[j - 1, j - 1]
                s_min_path[:j, j] = s_min_path[:j, j - 1]
                if self.prod.callput == CallPut.Call:
                    if self.prod.ave_method == AverageMethod.Arithmetic:
                        s_max_path[:j + 1, j] += np.maximum(s_path[:j + 1, j], self.limited_price)
                        s_min_path[:j + 1, j] += np.maximum(s_path[:j + 1, j], self.limited_price)
                    elif self.prod.ave_method == AverageMethod.Geometric:
                        s_max_path[:j + 1, j] += np.maximum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                        s_min_path[:j + 1, j] += np.maximum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                elif self.prod.callput == CallPut.Put:
                    if self.prod.ave_method == AverageMethod.Arithmetic:
                        s_max_path[:j + 1, j] += np.minimum(s_path[:j + 1, j], self.limited_price)
                        s_min_path[:j + 1, j] += np.minimum(s_path[:j + 1, j], self.limited_price)
                    elif self.prod.ave_method == AverageMethod.Geometric:
                        s_max_path[:j + 1, j] += np.minimum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))
                        s_min_path[:j + 1, j] += np.minimum(np.log(s_path[:j + 1, j]), np.log(self.limited_price))

        for i in range(obs_start_step, obs_end_step + 1):
            # 对加和除以天数，即为历史价格路径平均值的最大值/历史价格路径平均值的最小值
            s_min_path[:i + 1, i] = (s_min_path[:i + 1, i] / (i - obs_start_step + 1))
            s_max_path[:i + 1, i] = (s_max_path[:i + 1, i] / (i - obs_start_step + 1))

        if self.prod.ave_method == AverageMethod.Geometric:
            s_min_path = np.exp(s_min_path)
            s_max_path = np.exp(s_max_path)

        # 在历史路径平均股价的最大值和最小值之间，插入固定数量n_samples的数值点，s_path是一个三维数组
        s_path = np.linspace(s_min_path, s_max_path, n_samples)
        return s_path

    @staticmethod
    def _trans_v(ave_s, s_list, v_list):
        """线性插值，求出插入的标的价格数值点对应的期权价值
        Args:
            ave_s: np.ndarray，插入的数值点对应的标的价格
            s_list: np.ndarray，已知点对应的的标的价格
            v_list: np.ndarray，已知点对应的的期权价值
        Returns:
            ave_v: np.ndarray，插入的数值点对应的期权价值
        """
        ave_v = ave_s.copy()
        for i in range(len(s_list[0, :])):
            # np.interp(x, xp, fp)  x: 插值点 xp: 已知点的x坐标  fp: 已知点的y坐标
            ave_v[:, i] = np.interp(ave_s[:, i], s_list[:, i], v_list[:, i])
        return ave_v

    def cal_price(self, s_path, interp_s_path, obs_start_step, obs_end_step, n_samples):
        """递推，计算每个节点的每个可能的历史平均股价数值点对应的期权价格。
            考虑一个具体的平均股价数值点，该点在股价下一步上升或下降后分别对应于下一层一个节点的一个历史平均股价数值，
            但该价格不一定是在已计算出的数值点中，所以一般需要使用插值法求出数值点对应的期权价格。
            然后对股价上升或下降后对应数值所对应期权价格进行加权平均并贴现，即为当前层考虑节点的一个平均股价数值点对应的期权价格。
            不断重复上述过程，直到计算出根节点处期权的价格。
        Args:
            s_path: np.ndarray，上三角结构的标的资产价格CRR二叉树
            interp_s_path: np.ndarray，三维数组，其第一维数是n_samples，第二和第三维数等同于树s的第一维和第二维
                    相当于n_samples个二叉树，分别对应了历史路径平均值的极大值s_max_path与极小值s_min_path中间的一个数值点的二叉树
            obs_start_step: float，观察平均的开始时间点，对应的二叉树分支索引
            obs_end_step: float，观察平均的终止时间点，对应的二叉树分支索引
            n_samples: int，在每个树节点历史路径平均股价的最大值和最小值之间，插入数值点的数量
        Returns:
            float，期权现值
        """
        v_grid = np.zeros(shape=(n_samples, self.n_step + 1, self.n_step + 1))
        for j in range(self.n_step, -1, -1):
            # 终止条件
            if j == self.n_step:
                v_grid[:, :, j] = np.maximum(self.prod.callput.value * (interp_s_path[:, :, -1] - self.prod.strike), 0)
                continue
            if j < obs_start_step or j > obs_end_step:
                # 非观察期内，直接用下一步的上升和下降的v做加权平均并贴现
                v_grid[:, 0:j + 1, j] = (v_grid[:, 0:j + 1, j + 1] * self.p + v_grid[:, 1:j + 2, j + 1] * (1 - self.p)
                                         ) * math.exp(-self.r * self.dt)
            else:  # 观察期内
                # 考虑某一个平均股价数值点，用插值法找到该点在股价下一步上升或下降后的期权价格，然后加权平均并贴现
                up_ave_s = ((j - obs_start_step + 1) * interp_s_path[:, 0:j + 1, j] + s_path[0:j + 1, j + 1]) / (
                        j - obs_start_step + 2)
                down_ave_s = ((j - obs_start_step + 1) * interp_s_path[:, 0:j + 1, j] + s_path[1:j + 2, j + 1]) / (
                        j - obs_start_step + 2)
                v_grid[:, 0:j + 1, j] = self._trans_v(up_ave_s, interp_s_path[:, 0:j + 1, j + 1],
                                                      v_grid[:, 0:j + 1, j + 1]) * self.p
                v_grid[:, 0:j + 1, j] += self._trans_v(down_ave_s, interp_s_path[:, 1:j + 2, j + 1],
                                                       v_grid[:, 1:j + 2, j + 1]) * (1 - self.p)
                v_grid[:, 0:j + 1, j] *= math.exp(-self.r * self.dt)
        return np.mean(v_grid[:, 0, 0])

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert (prod.ave_method == AverageMethod.Arithmetic and prod.substitute == AsianAveSubstitution.Underlying
                and not prod.enhanced), "仅支持算术平均代替资产结算价的亚式期权，不支持几何平均、替代执行价、增强亚式"
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date) / prod.t_step_per_year
        obs_start = prod.trade_calendar.business_days_between(calculate_date, prod.obs_start) / prod.t_step_per_year
        obs_end = prod.trade_calendar.business_days_between(calculate_date, prod.obs_end) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        if prod.enhanced:
            self.limited_price = prod.limited_price
        else:
            if prod.callput == CallPut.Call:
                self.limited_price = 0.0001
            else:  # prod.callput == CallPut.Put:
                self.limited_price = 100 * prod.strike  # np.inf

        self.n_step = int(_maturity_business_days * self.tree_branches)
        self.r = self.process.interest(_maturity)
        self.q = self.process.div(_maturity)
        self.vol = self.process.vol(_maturity_business_days, spot)
        # 生成标的价格树
        s_path = self.path_generator(self.n_step, _maturity_business_days, spot).copy()
        obs_start_step = int(obs_start * self.tree_branches)
        obs_end_step = int(obs_end * self.tree_branches)
        # 根据标的价格二叉树，计算每个树节点对应的历史路径平均值的极大值s_max_path与极小值s_min_path，在其中插入固定数量n_samples的数值点
        interp_s_path = self.cal_path(s_path, obs_start_step, obs_end_step, n_samples=self.n_samples)
        # 递推，计算每个节点的每个可能的历史平均股价数值点对应的期权价格
        price = self.cal_price(s_path, interp_s_path, obs_start_step, obs_end_step, n_samples=self.n_samples)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from functools import lru_cache
import numpy as np
from pricelib.common.utilities.enums import UpDown, InOut, PaymentType
from pricelib.common.utilities.patterns import HashableArray
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import QuadEngine


class QuadBarrierEngine(QuadEngine):
    """障碍期权数值积分定价引擎
    只支持离散观察(默认为每日观察)；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付"""

    @lru_cache(maxsize=1)
    def set_quad_price_range(self, s_sliced, barrier, updown):
        """
        设置数值积分的区间(价格范围)
        Args:
            s_sliced: np.ndarray = self.s[:, step], 其中step是从后向前的步数
            barrier: float, 障碍价格
            updown: 向上/向下，UpDown枚举类
        Returns:
            quad_range: tuple, (np.ndarray, ), 数值积分的区间(索引)
            an_range: tuple, (np.ndarray, ), 解析式Vma、Vmb计算积分的区间(索引)
        """
        if updown == UpDown.Up:
            quad_range = np.where(s_sliced < barrier)
            an_range = np.where(s_sliced >= barrier)
        elif updown == UpDown.Down:
            quad_range = np.where(s_sliced > barrier)
            an_range = np.where(s_sliced <= barrier)
        else:
            raise ValueError("updown must be Up or Down")
        return quad_range, an_range

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        dt = prod.discrete_obs_interval
        self.backward_steps = int(tau / dt)
        self._check_method_params()

        # v_grid = np.empty(shape=(self.n_points, self.backward_steps + 1))
        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)
        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, tau)
        s_vec = np.exp(self.ln_s_vec)

        # upper_barrier = self.n_max * spot
        # lower_barrier = 1
        # s_vec = np.linspace(lower_barrier, upper_barrier, self.n_points)
        # s_vec = HashableArray(np.tile(s_vec, reps=(self.backward_steps + 1, 1)).T)

        # 从后向前进行积分计算
        # 设定期末价值，在障碍价格内侧，默认设定未敲入，未敲出
        if prod.inout == InOut.Out:
            v_vec = np.maximum(prod.callput.value * (s_vec - prod.strike), 0) * prod.parti
            if prod.updown == UpDown.Up:
                v_vec[np.where(s_vec > prod.barrier)] = prod.rebate
            elif prod.updown == UpDown.Down:
                v_vec[np.where(s_vec < prod.barrier)] = prod.rebate
            else:
                raise ValueError("updown must be Up or Down")
        elif prod.inout == InOut.In:
            v_vec = prod.rebate * np.ones(len(s_vec))
            if prod.updown == UpDown.Up:
                upper_index = np.where(s_vec > prod.barrier)
                v_vec[upper_index] = np.maximum(prod.callput.value * (s_vec[upper_index] - prod.strike),
                                                0) * prod.parti
            elif prod.updown == UpDown.Down:
                lower_index = np.where(s_vec < prod.barrier)
                v_vec[lower_index] = np.maximum(prod.callput.value * (s_vec[lower_index] - prod.strike),
                                                0) * prod.parti
            else:
                raise ValueError("updown must be Up or Down")
        else:
            raise ValueError("inout must be In or Out")

        for step in range(self.backward_steps - 1, 0, -1):
            quad_range, an_range = self.set_quad_price_range(s_vec, prod.barrier, prod.updown)
            # 数值积分计算区域
            y = self.ln_s_vec
            x = self.ln_s_vec[quad_range]
            v_vec[quad_range] = self.fft_step_backward(x, y, v_vec, dt)
            # v_grid[:, step] = np.minimum(v_grid[:, step], 1e250)  # 防止价格极大和极小时，期权价值趋近于无穷大
            # 解析公式计算积分区域
            if prod.inout == InOut.In:
                x = HashableArray(self.ln_s_vec[an_range])
                v_vec[an_range] = self.Vma(np.exp(x), prod.strike, prod.callput.value,
                                           dt * (self.backward_steps - step)) * prod.callput.value * prod.parti
                v_vec[an_range] -= self.Vmb(np.exp(x), prod.strike, prod.callput.value, dt * (
                        self.backward_steps - step)) * prod.strike * prod.callput.value * prod.parti
            elif prod.inout == InOut.Out:
                if prod.payment_type == PaymentType.Expire:  # 到期支付现金补偿
                    v_vec[an_range] = prod.rebate * np.exp(-r * dt * (self.backward_steps - step))
                else:  # 立即支付现金补偿
                    v_vec[an_range] = prod.rebate
            else:
                raise ValueError("inout must be In or Out")

        x = np.log(np.array([spot]))
        value = self.fft_step_backward(x, self.ln_s_vec, v_vec, dt)[0]
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import InOut, UpDown, PaymentType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import BiTreeEngine
from pricelib.common.utilities.patterns import HashableArray


class BiTreeBarrierEngine(BiTreeEngine):
    """障碍期权二叉树定价引擎
    只支持离散观察(默认为每日观察)；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        self.r = self.process.interest(tau)
        self.q = self.process.div(tau)
        self.vol = self.process.vol(tau, spot)

        n_step = round(tau * self.tree_branches)
        s_paths = self.path_generator(n_step, tau, spot).copy()
        v_grid = np.zeros(shape=s_paths.shape)

        # 设置观察时点obs_points
        if prod.discrete_obs_interval is None:
            obs_points = np.arange(0, n_step + 1, 1)[1:]
        else:
            obs_points = np.flip(np.round(np.arange(tau, 0, -prod.discrete_obs_interval) * self.tree_branches))
            obs_points = np.concatenate((np.array([0]), obs_points))
        obs_points = np.round(obs_points).astype(int)

        for j in range(n_step, -1, -1):
            if j == n_step:  # 到期日，设置终止条件
                if prod.inout == InOut.Out:
                    v_grid[:, -1] = np.maximum(prod.callput.value * (s_paths[:, -1] - prod.strike), 0) * prod.parti
                    if prod.updown == UpDown.Up:
                        v_grid[np.where(s_paths[:, -1] > prod.barrier), -1] = prod.rebate
                    elif prod.updown == UpDown.Down:
                        v_grid[np.where(s_paths[:, -1] < prod.barrier), -1] = prod.rebate
                    else:
                        raise ValueError("updown must be Up or Down")
                elif prod.inout == InOut.In:
                    v_grid[:, -1] = prod.rebate
                    if prod.updown == UpDown.Up:
                        upper_index = np.where(s_paths[:, -1] > prod.barrier)
                        v_grid[upper_index, -1] = np.maximum(
                            prod.callput.value * (s_paths[upper_index, -1] - prod.strike), 0) * prod.parti
                    elif prod.updown == UpDown.Down:
                        lower_index = np.where(s_paths[:, -1] < prod.barrier)
                        v_grid[lower_index, -1] = np.maximum(
                            prod.callput.value * (s_paths[lower_index, -1] - prod.strike), 0) * prod.parti
                    else:
                        raise ValueError("updown must be Up or Down")
            else:
                v_grid[:-1, j] = v_grid[:-1, j + 1] * self.p + v_grid[1:, j + 1] * (1 - self.p)
                if j in obs_points:
                    time_remain = self.dt * (n_step - j)
                    if prod.updown == UpDown.Up:
                        knock_inout = np.where(s_paths[:j, j] >= prod.barrier)
                    elif prod.updown == UpDown.Down:
                        knock_inout = np.where(s_paths[:j, j] <= prod.barrier)
                    else:
                        raise ValueError("updown must be Up or Down")
                    if prod.inout == InOut.In:
                        x = HashableArray(s_paths[knock_inout, j])
                        v_grid[knock_inout, j] = self.Vma(x, prod.strike, prod.callput.value,
                                                          time_remain) * prod.callput.value * prod.parti
                        v_grid[knock_inout, j] -= self.Vmb(x, prod.strike, prod.callput.value,
                                                           time_remain) * prod.strike * prod.callput.value * prod.parti
                    elif prod.inout == InOut.Out:
                        if prod.payment_type == PaymentType.Expire:  # 到期支付现金补偿
                            v_grid[knock_inout, j] = prod.rebate * np.exp(-self.r * time_remain)
                        else:  # 立即支付现金补偿
                            v_grid[knock_inout, j] = prod.rebate
                    else:
                        raise ValueError("inout must be In or Out")
                v_grid[:-1, j] *= np.exp(-self.r * self.dt)
        return v_grid[0, 0]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import McEngine


class MCRangeAccuralEngine(McEngine):
    """区间累计 Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        # 敲出与行权日期
        n_coupon = np.sum((paths < prod.upper_strike) & (paths > prod.lower_strike))

        payoff = self.process.interest.disc_factor(_maturity) * prod.payment * n_coupon / prod.t_step_per_year
        return payoff * prod.s0 / self.n_path



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from numba import njit
from ..utilities.enums import ProcessType
from ..utilities.patterns import Observable, Observer
from .stoch_process import StochProcessBase


@njit(parallel=True, fastmath=True, cache=True)
def evolve_bs(t0, x0, dt, dw, drift, diffusion):
    """Black-Scholes-Merton SDE演化函数，jit加速
    Black-Scholes-Merton SDE: dS/S = drift * dt + diffusion * dW
    Args:
        t0: float, 距离起始日的年化时间
        x0: np.ndarray, t0时刻的标的价格向量
        dt: float, 年化时间增量
        dw: np.ndarray, 布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        drift: float, BSM的漂移率
        diffusion: np.ndarray, BSM的波动率
    Returns:
        np.ndarray, t0+dt时刻的标的价格向量
    """
    return x0 * (1 + drift * dt + diffusion * dw * np.sqrt(dt))


class GeneralizedBSMProcess(StochProcessBase, Observer, Observable):
    """BS风险中性 SDE: dS/S = (r-q)dt + vol dW
    既是观察者又是被观察者，观察者模式的中介者。观察S、r、q、sigma的变化，被定价引擎engine观察。
    """
    process_type = ProcessType.BSProcess1D

    def __init__(self, spot: Observable, interest: Observable, div: Observable, vol: Observable):
        """标的价格、无风险利率、分红融券率、波动率，都是被观察者。
        vol可以是常数波动率，也可以是局部波动率，可以由vol(t,S)取值"""
        self.spot = spot
        self.interest = interest
        self.div = div
        self.vol = vol
        spot.add_observer(self)  # 将本实例加入spot观察者列表
        interest.add_observer(self)  # 将本实例加入interest观察者列表
        div.add_observer(self)  # 将本实例加入div观察者列表
        vol.add_observer(self)  # 将本实例加入vol观察者列表
        super().__init__()  # 调用被观察者Observable的构造函数

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.spot.remove_observer(self)
        self.interest.remove_observer(self)
        self.div.remove_observer(self)
        self.vol.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """被观察者的值发生变化时，自动调用update方法，通知观察者"""
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    def __call__(self):
        return ProcessType.BSProcess1D

    def drift(self, t):
        """BSM过程的漂移项
        Args:
            t: float，距离起始日的年化时间
        Returns:
            drift: float = r - q，r是无风险利率，q是分红融券率
        """
        return self.interest(t) - self.div(t)

    def diffusion(self, t, spot):
        """BSM过程的扩散项
        Args:
            t: float，距离起始日的年化时间
            spot: np.ndarray，标的价格向量
        Returns:
            diffusion: np.ndarray = vol，vol是波动率
        """
        return self.vol(t, spot)

    def evolve(self, t, x, dt, dw):
        """演化函数，根据BSM SDE进行演化
        Args:
            t: float，距离起始日的年化时间
            x: np.ndarray, t0时刻的标的价格向量
            dt: float, 年化时间增量
            dw: np.ndarray, 布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        Returns:
            np.ndarray, t0+dt时刻的标的价格向量
        """
        return evolve_bs(t, x, dt, dw, self.drift(t), self.diffusion(t, x))

    def get_fn_pde_coef(self, maturity, spot):
        """获取返回pde系数的函数
        Args:
            maturity: float，年化到期时间
            spot: float，标的价格
        Returns:
            fn_pde_coef: function, 返回PDE系数组件abc的函数
        """

        # TODO: 目前局部波动率使用K坐标，以后可能重构为log moneyness坐标
        # def fn_pde_coef(u, z):  # z在校准中是执行价K，在定价中是标的价S_u, u是距离起始日的年化时间
        #     """返回PDE系数组件abc的函数，设pde有限差分的价格向量索引为i_vec，
        #     a是需要乘以i_vec^2的系数，b是需要乘以i_vec的系数，c是不需要乘以i_vec的系数"""
        #     r_u = self.interest(u)  # u时刻的无风险利率r
        #     q_u = self.div(u)  # u时刻的分红融券率q
        #     fwd_u = spot * math.exp((maturity - u) * (r_u - q_u))  # u时刻的远期价格
        #     lv = self.vol(u, np.log(z / fwd_u))  # u时刻，log(K/F)对数moneyness的local_vol
        #     a, b, c = lv ** 2, r_u - q_u, r_u
        #     return a, b, c
        def fn_pde_coef(t, spot):  # z在校准中是执行价K，在定价中是标的价S_u, u是距离起始日的年化时间
            """返回PDE系数组件abc的函数
            Args:
                t: float，年化到期时间
                spot: float，标的价格
            Returns: 设pde有限差分的价格向量索引为i_vec
                a: 需要乘以i_vec^2的系数
                b: 需要乘以i_vec的系数
                c: 不需要乘以i_vec的系数
            """
            r_t = self.interest(t)  # t时刻的无风险利率r
            q_t = self.div(t)  # t时刻的分红融券率q
            lv = self.vol(t, spot)  # t时刻，spot的local_vol
            a, b, c = lv ** 2, r_t - q_t, r_t
            return a, b, c

        return fn_pde_coef



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType, QuadMethod
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.time import global_evaluation_date


class QuadSnowballEngine(QuadEngine):
    """雪球数值积分法定价引擎
    只支持不是仅到期观察敲入的结构 (只有到期观察敲入的结构，如FCN、DCN等另有更快的算法)
    支持变敲出、变敲入、变票息等要素可变型雪球结构"""

    def __init__(self, stoch_process=None, quad_method=QuadMethod.Simpson, n_points=1301, *,
                 s=None, r=None, q=None, vol=None):
        """初始化数值积分法定价引擎
        Args:
            stoch_process: StochProcessBase，随机过程
            quad_method: 数值积分方法，QuadMethod枚举类，Trapezoid梯形法则/Simpson辛普森法则
            n_points: int，积分点个数 (需要注意，辛普森法则的n_points必须是奇数)
            n_max: int，价格网格上界，设为n_smax倍初始价格
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, quad_method, n_points, s=s, r=r, q=q, vol=vol)
        # 以下为计算过程的中间变量
        self.prod = None  # Product产品对象
        self.obs_dates = None  # 根据估值日，将敲出观察日转化为List[int]，交易日期限
        self.pay_dates = None  # 根据估值日，将支付日转化为List[float]，年化自然日期限
        self.diff_obs_pay_dates = None  # 观察日与付息日的差值，List[float]，年化自然日期限
        self.smax = None  # 价格网格上界，float
        self.smin = None  # 价格网格下界，float
        self.s_vec = None  # 价格格点(不含smax和smin)，np.ndarray
        self.dt = None  # 时间步长，float
        self.j_vec = None  # 时间格点向量，np.ndarray
        self.v_not_in = None  # 未敲入时的期权价值矩阵，np.ndarray
        self.v_knock_in = None  # 敲入时的期权价值矩阵，np.ndarray
        self.next_paydate = None
        self.out_dates = None  # 逆序敲出观察日，List[int]，交易日期限
        self.next_diff_obspaydate = None  # 每个交易日对应的下一个敲出观察日对应的付息日与观察日的自然日年化之差
        self.next_barrier_out = None
        self.next_barrier_in = None
        self.out_idxs = None
        self.in_idx = None
        self.itm = None
        self.next_coupon_out = None
        self.reversed_barrier_out = None
        self.reversed_barrier_in = None
        self.reversed_coupon_out = None
        self.coupon_outs = None
        self._barrier_in = None  # 是经过估值日截断之后的列表，例如prod.barrier_in有22个，存续一年时估值，self._barrier_in只有12个
        self._barrier_out = None  # 是经过估值日截断之后的列表，例如prod.barrier_out有22个，存续一年时估值，self._barrier_out只有12个
        self._coupon_out = None  # 是经过估值日截断之后的列表，例如prod.coupon_out有22个，存续一年时估值，self._coupon_out只有12个

    # pylint: disable=too-many-locals, too-many-statements
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        self.backward_steps = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        obs_dates = np.array([num for num in obs_dates if num >= 0])
        calculate_start_diff = (calculate_date - prod.start_date).days
        pay_dates = prod.pay_dates.count_calendar_days(prod.start_date)
        self.pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= calculate_start_diff])
        assert len(obs_dates) == len(self.pay_dates), f"Error: {prod}的观察日和付息日长度不一致"
        self.diff_obs_pay_dates = np.array([(prod.pay_dates.date_schedule[i] - prod.obs_dates.date_schedule[i]).days
                                            for i in range(len(obs_dates))]) / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(_maturity)
        q = self.process.div(_maturity)
        vol = self.process.vol(_maturity, spot)
        self._check_method_params()
        self.set_quad_params(r=r, q=q, vol=vol)
        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, self.backward_steps / prod.t_step_per_year)

        self.s_vec = np.exp(self.ln_s_vec)
        self.dt = 1 / prod.t_step_per_year
        self.j_vec = np.linspace(0, self.backward_steps, self.backward_steps + 1)  # 生成时间格点
        self.v_not_in = np.zeros(shape=(self.n_points, self.backward_steps + 1))
        self.v_knock_in = np.zeros(shape=(self.n_points, self.backward_steps + 1))

        # 初始化逆序敲出观察日
        if isinstance(obs_dates, (list, np.ndarray)):  # 敲出观察日是具体第n天数的列表，如[11,22,32,43,53,64]
            self.out_dates = np.round(np.flip(np.array(obs_dates))).astype(int)
        elif isinstance(obs_dates, (int, np.int32)):  # 敲出观察日是整数， 敲出观察日个数，如6代表6个均匀分布的观察日
            out_dates = (self.backward_steps - np.round(np.linspace(0, self.backward_steps,
                                                                    round(obs_dates + 1)))[:-1])
            self.out_dates = out_dates.astype(int)
        else:
            raise ValueError('敲出观察日类型设置错误，请检查')
        self.next_diff_obspaydate = self.diff_obs_pay_dates.repeat(
            np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))  # 计息时间数
        self.next_diff_obspaydate = np.append(self.next_diff_obspaydate[0],
                                              self.next_diff_obspaydate)  # 每个交易日对应的下一个敲出观察日对应的付息日与观察日的自然日年化之差
        # 初始化票息支付日
        self.next_paydate = self.pay_dates.repeat(
            np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))  # 计息时间数
        self.next_paydate = np.append(self.pay_dates[0], self.next_paydate)  # 每个交易日对应的下一个敲出观察日对应的付息日

        # 设置期末价值状态
        self._init_terminal_condition(self.s_vec, _maturity)

        if 0 in self.out_dates:
            # 如果估值日就是敲出观察日
            j = 0
            # 如果敲出价是浮动的，调整敲出价
            start_barrier_out = (self.reversed_barrier_out[np.where(self.out_dates == j)[0][0]]
                                     if isinstance(prod.barrier_out, (list, np.ndarray)) else prod.barrier_out)
            if spot >= start_barrier_out:  # 发生敲出
                # 如果敲出票息是浮动的，调整敲出票息
                start_coupon_out = (self.reversed_coupon_out[np.where(self.out_dates == j)[0][0]]
                                        if isinstance(prod.coupon_out, (list, np.ndarray)) else prod.coupon_out)
                start_t = 1 if prod.trigger else self.next_paydate[j]
                return np.where(spot - prod.strike_call > 0, (spot - prod.strike_call) * prod.parti_out, 0) + (
                        prod.margin_lvl + start_coupon_out * start_t) * prod.s0 * self.process.interest.disc_factor(
                    self.next_paydate[j], self.next_paydate[j] - self.next_diff_obspaydate[j])

        for j in range(self.backward_steps - 1, 0, -1):
            # 未敲入网格更新
            if prod.status == StatusType.NoTouch:
                self.v_not_in[:, j] = self.fft_step_backward(self.ln_s_vec, self.ln_s_vec,
                                                             self.v_not_in[:, j + 1], self.dt)
            # 已敲入网格更新
            self.v_knock_in[:, j] = self.fft_step_backward(self.ln_s_vec, self.ln_s_vec,
                                                           self.v_knock_in[:, j + 1], self.dt)

            if j in self.out_dates:
                coupon_t = 1 if prod.trigger else self.next_paydate[j]
                # 如果敲出价是浮动的，调整敲出价
                self.next_barrier_out = (self.reversed_barrier_out[np.where(self.out_dates == j)[0][0]]
                                         if isinstance(prod.barrier_out, (list, np.ndarray)) else prod.barrier_out)
                self.out_idxs = np.array(np.where(self.s_vec >= self.next_barrier_out)[0])
                # 如果敲出票息是浮动的，调整敲出票息
                self.next_coupon_out = (self.reversed_coupon_out[np.where(self.out_dates == j)[0][0]]
                                        if isinstance(prod.coupon_out, (list, np.ndarray)) else prod.coupon_out)
                # 发生敲出的payoff
                knock_out_payoff = self.itm[self.out_idxs] + (
                    prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0 * self.process.interest.disc_factor(
                    self.next_paydate[j], self.next_paydate[j] - self.next_diff_obspaydate[j])
                # 用knock_out_payoff覆盖敲出价上方的衍生品价值
                self.v_not_in[self.out_idxs, j] = self.v_knock_in[self.out_idxs, j] = knock_out_payoff

                # 如果敲入价是浮动的，调整敲入价
                self.next_barrier_in = self.reversed_barrier_in[np.where(self.out_dates == j)[0][0]] if isinstance(
                    prod.barrier_in, (list, np.ndarray)) else prod.barrier_in
                self.in_idx = np.where(self.s_vec <= self.next_barrier_in)[0][
                    -1] if self.next_barrier_in > 0 else 0

            if prod.status == StatusType.NoTouch:
                # 未敲入网格-敲入线下方的价值等于已敲入网格
                self.v_not_in[:self.in_idx, j] = self.v_knock_in[:self.in_idx, j]

        x = np.array([np.log(spot)])
        if prod.status == StatusType.NoTouch:
            value = self.fft_step_backward(x, self.ln_s_vec, self.v_not_in[:, 1], self.dt)[0]
        else:
            value = self.fft_step_backward(x, self.ln_s_vec, self.v_knock_in[:, 1], self.dt)[0]
        return value

    def _init_terminal_condition(self, s_vec, maturity):
        """初始化终止时已敲入且未敲出、未敲入且未敲出的期权价值
        Args:
            s_vec: List[float], 网格的价格格点
            maturity: float, 到期时间，年化自然日期限
        Returns: None
        """
        prod = self.prod
        # 初始化next_barrier_in、next_barrier_out、next_coupon_out为最后一期的敲入价、敲出价、敲出票息。若变敲出变敲入，需要动态调整
        # 设置到期日的敲入边界位置
        if isinstance(prod.barrier_in, (int, float, np.int32, np.float64)):
            # 常数敲入雪球
            self.next_barrier_in = prod.barrier_in
        elif isinstance(prod.barrier_in, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_in = prod.barrier_in[-len(self.out_dates):].copy()
            # 浮动敲入雪球
            self.next_barrier_in = self._barrier_in[-1]
            self.reversed_barrier_in = np.flip(np.array(self._barrier_in))
        else:
            raise ValueError(f'敲入障碍价类型为{type(prod.barrier_in)}，仅支持int/float/list/np.ndarray，请检查')

        # 设置到期日的敲出边界位置
        if isinstance(prod.barrier_out, (int, float, np.int32, np.float64)):
            # 常数敲出雪球
            self.next_barrier_out = prod.barrier_out
        elif isinstance(prod.barrier_out, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_out = prod.barrier_out[-len(self.out_dates):].copy()
            # 阶梯式雪球
            self.next_barrier_out = self._barrier_out[-1]
            self.reversed_barrier_out = np.flip(np.array(self._barrier_out))
        else:
            raise ValueError(f'敲出障碍价类型为{type(prod.barrier_out)}，仅支持int/float/list/np.ndarray，请检查')

        # 敲出票息
        if isinstance(prod.coupon_out, (int, float, np.int32, np.float64)):
            # 常数敲出票息
            self.next_coupon_out = prod.coupon_out
            # N天的敲出票息列表，用于生成mspot价格上边界条件
            self.coupon_outs = np.full(self.backward_steps + 1, prod.coupon_out)
        elif isinstance(prod.coupon_out, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._coupon_out = prod.coupon_out[-len(self.out_dates):].copy()
            # 浮动敲出票息
            self.next_coupon_out = self._coupon_out[-1]
            self.reversed_coupon_out = np.flip(np.array(self._coupon_out))
            # N天的敲出票息列表，用于生成smax价格上边界条件
            coupon_outs = np.array(self._coupon_out).repeat(
                np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))
            self.coupon_outs = np.append(coupon_outs[0], coupon_outs)
        else:
            raise ValueError(f'敲出票息类型为{type(prod.coupon_out)}，仅支持int/float/list/np.ndarray，请检查')

        # in_idx：小于到期日敲入线的最大的标的价格索引。如果是变敲入雪球，此变量in_idx应该随敲入价动态调整。
        if self.next_barrier_in > 0 and s_vec[0] < self.next_barrier_in:
            self.in_idx = np.where(s_vec <= self.next_barrier_in)[0][-1]
        else:
            self.in_idx = 0
        # out_idxs：大于到期日敲出线的标的价格数组的索引。如果是变敲出雪球，此变量out_idxs应该随敲出价动态调整。
        self.out_idxs = np.array(np.where(s_vec >= self.next_barrier_out)[0])
        # 虚值call在值度
        self.itm = np.where(s_vec - prod.strike_call > 0, (s_vec - prod.strike_call) * prod.parti_out, 0)

        coupon_t = 1 if prod.trigger else self.pay_dates[-1]  # 票息是否年化
        # 矩阵最后一列：到期payoff。考虑到了保底、保证金\纯期权模式、以及敲出边界上方增加虚值call
        """已敲入：如果到期小于敲出边界，在100%本金模式下，小于保底边界，获得保底价；
                                                  大于保底边界，获得到期价格；
                                   在保证金模式下，小于保底边界，支付期初价（高行权价）与保底价之间价差；
                                                  大于保底边界，支付到期价格与期初价之间价差；
                  如果到期大于敲出边界，大于看涨行权价，获得看涨收益、本金和票息；
                                    在敲出边界和看涨行权价之间，获得本金和票息"""
        self.v_knock_in[:, -1] = np.where(s_vec < self.next_barrier_out,
                                             np.minimum(
                                                 np.where(s_vec <= prod.strike_lower, prod.strike_lower,
                                                          s_vec) -
                                                 prod.strike_upper, 0) + prod.s0 * prod.margin_lvl,
                                             np.where(s_vec > prod.strike_call,
                                                      (s_vec - prod.strike_call) * prod.parti_out + (
                                                          prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0,
                                                      (
                                                          prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0))
        """未敲入：如果到期小于敲入边界，在100%本金模式下，小于保底边界，获得保底价；
                                                  大于保底边界，获得实际价格；
                                  在保证金模式下，小于保底边界，支付保底价与期初价之间价差；
                                                  大于保底边界，支付到期价格与期初价之间价差；
                 如果到期大于敲入边界，且大于看涨行权价，获得看涨收益、本金和票息；
                                  在敲入边界和看涨行权价之间，获得本金和票息
                                          如果在敲出边界和看涨行权价之间，获得本金和敲出票息
                                          如果在敲入边界和敲出边界之间，获得本金和红利票息"""
        self.v_not_in[:, -1] = np.where(s_vec <= self.next_barrier_in,
                                           np.where(s_vec <= prod.strike_lower, prod.strike_lower, s_vec) -
                                           prod.strike_upper + prod.s0 * prod.margin_lvl,
                                           np.where(s_vec > prod.strike_call,
                                                    (s_vec - prod.strike_call) * prod.parti_out + (
                                                        prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0,
                                                    np.where(s_vec >= self.next_barrier_out,
                                                             (
                                                         prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0,
                                                             (
                                                         prod.margin_lvl + prod.coupon_div * coupon_t) * prod.s0)))



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from scipy.stats import norm
from pricelib.common.utilities.enums import UpDown, BarrierType, PaymentType, InOut
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticBarrierEngine(AnalyticEngine):
    """障碍期权闭式解定价引擎
    敲入期权现在支付期权费，但是当到期前资产价格触及障碍水平时，期权才生效。若一直没有发生敲入，则到期时支付现金返还rebate
    敲出期权现在支付期权费，但是当到期前资产价格触及障碍水平时，期权就失效了。若到期前发生敲出事件，则立刻支付现金返还rebate
    Merton(1973), Reiner & Rubinstein(1991a)提出障碍期权解析解，
    Broadie, Glasserman和Kou(1995)提出均匀离散观察障碍期权近似解"""

    # pylint: disable=invalid-name, too-many-locals, missing-docstring, too-many-branches, too-many-statements
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        if prod.inout == InOut.In:
            assert prod.payment_type == PaymentType.Expire, "ValueError: 敲入期权解析解，一直未敲入，到期时支付现金返还。payment_type应该是Expire"
        if prod.inout == InOut.Out:
            assert prod.payment_type == PaymentType.Hit, "ValueError: 敲出期权解析解，发生敲出，立刻支付现金返还。payment_type应该是Hit"
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, self.process.spot())
        drift = r - q
        if prod.discrete_obs_interval is not None:
            # 均匀离散观察障碍期权，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
            # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
            if prod.updown == UpDown.Up:  # 向上
                barrier = prod.barrier * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            elif prod.updown == UpDown.Down:  # 向下
                barrier = prod.barrier * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            else:
                raise ValueError("self.updown 只能是Up或Down")
        else:  # 连续观察障碍期权
            barrier = prod.barrier

        mu = drift / vol ** 2 - 0.5
        la = np.sqrt(mu ** 2 + 2 * r / vol ** 2)
        a = (barrier / spot) ** (2 * mu)
        b = (barrier / spot) ** (2 * mu + 2)
        c = (barrier / spot) ** (mu + la)
        d = (barrier / spot) ** (mu - la)
        a1 = np.log(spot / prod.strike)
        a2 = np.log(spot / barrier)
        a3 = np.log(spot * prod.strike / barrier ** 2)
        a4 = drift + 0.5 * vol ** 2
        a5 = drift - 0.5 * vol ** 2
        a6 = tau
        a7 = vol * np.sqrt(a6)
        d1 = (a1 + a4 * a6) / a7
        d2 = (a1 + a5 * a6) / a7
        d3 = (a2 + a4 * a6) / a7
        d4 = (a2 + a5 * a6) / a7
        d5 = (a2 - a5 * a6) / a7
        d6 = (a2 - a4 * a6) / a7
        d7 = (a3 - a5 * a6) / a7
        d8 = (a3 - a4 * a6) / a7
        d9 = -a2 / a7 + la * a7
        d10 = -a2 / a7 - la * a7

        def A(phi):
            return phi * spot * np.exp(-q * a6) * norm.cdf(phi * d1) - \
                phi * prod.strike * np.exp(-r * a6) * norm.cdf(phi * d2)

        def B(phi):
            return phi * spot * np.exp(-q * a6) * norm.cdf(phi * d3) - \
                phi * prod.strike * np.exp(-r * a6) * norm.cdf(phi * d4)

        def C(phi, eta):
            return phi * spot * np.exp(-q * a6) * b * norm.cdf(-eta * d8) - \
                phi * prod.strike * np.exp(-r * a6) * a * norm.cdf(-eta * d7)

        def D(phi, eta):
            return phi * spot * np.exp(-q * a6) * b * norm.cdf(-eta * d6) - \
                phi * prod.strike * np.exp(-r * a6) * a * norm.cdf(-eta * d5)

        def E(eta):
            return prod.rebate * np.exp(-r * a6) * (norm.cdf(eta * d4) - a * norm.cdf(-eta * d5))

        def F(eta):
            return prod.rebate * (c * norm.cdf(eta * d9) + d * norm.cdf(eta * d10))

        if prod.barrier_type == BarrierType.UIC:  # 向上敲入看涨
            phi = 1
            eta = -1
            if prod.strike >= barrier:
                price = A(phi) + E(eta)
            else:
                price = (B(phi) - C(phi, eta) + D(phi, eta)) + E(eta)
        elif prod.barrier_type == BarrierType.UIP:  # 向上敲入看跌
            phi = -1
            eta = -1
            if prod.strike >= barrier:
                price = (A(phi) - B(phi) + D(phi, eta)) + E(eta)
            else:
                price = C(phi, eta) + E(eta)
        elif prod.barrier_type == BarrierType.UOC:  # 向上敲出看涨
            phi = 1
            eta = -1
            if prod.strike >= barrier:
                price = F(eta)
            else:
                price = (A(phi) - B(phi) + C(phi, eta) - D(phi, eta)) + F(eta)
        elif prod.barrier_type == BarrierType.UOP:  # 向上敲出看跌
            phi = -1
            eta = -1
            if prod.strike >= barrier:
                price = (B(phi) - D(phi, eta)) + F(eta)
            else:
                price = (A(phi) - C(phi, eta)) + F(eta)
        elif prod.barrier_type == BarrierType.DIC:  # 向下敲入看涨
            phi = 1
            eta = 1
            if prod.strike >= barrier:
                price = C(phi, eta) + E(eta)
            else:
                price = (A(phi) - B(phi) + D(phi, eta)) + E(eta)
        elif prod.barrier_type == BarrierType.DIP:  # 向下敲入看跌
            phi = -1
            eta = 1
            if prod.strike >= barrier:
                price = (B(phi) - C(phi, eta) + D(phi, eta)) + E(eta)
            else:
                price = A(phi) + E(eta)
        elif prod.barrier_type == BarrierType.DOC:  # 向下敲出看涨
            phi = 1

            eta = 1
            if prod.strike >= barrier:
                price = (A(phi) - C(phi, eta)) + F(eta)
            else:
                price = (B(phi) - D(phi, eta)) + F(eta)
        elif prod.barrier_type == BarrierType.DOP:  # 向下敲出看跌
            phi = -1
            eta = 1
            if prod.strike >= barrier:
                price = (A(phi) - B(phi) + C(phi, eta) - D(phi, eta)) + F(eta)
            else:
                price = F(eta)
        else:
            raise ValueError("self.barrier_type 障碍类型错误")
        return price * prod.parti



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.utilities.enums import CallPut, InOut, UpDown, BarrierType, PaymentType, EngineType, StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticBarrierEngine


class BarrierOption(OptionBase, Observer):
    """单边障碍期权
    敲入期权在障碍被触及时才会生效，敲出期权在障碍被触及时立即失效"""

    def __init__(self, strike, barrier, rebate, callput: CallPut, inout: InOut, updown: UpDown, *, parti=1,
                 payment_type=None, engine=None, status=StatusType.NoTouch, maturity=None,
                 start_date=None, end_date=None, discrete_obs_interval=None, trade_calendar=CN_CALENDAR,
                 annual_days=AnnualDays.N365, t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            strike: float, 执行价
            barrier: float, 障碍价
            rebate: float, 敲出现金返还金额，或未敲入现金返还金额，绝对数值，非百分比，非年化
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            inout: 敲入敲出类型，InOut枚举类，In敲入/Out敲出
            updown: 向上向下类型，UpDown枚举类，Up/Down
            payment_type: 现金返还支付类型，PaymentType枚举类，敲入默认为到期支付Expire；敲出默认为立即支付Hit
            parti: float, 香草期权的参与率，默认为1
            status: 估值日前的敲入敲出状态，StatusType枚举类，默认为NoTouch未敲入/未敲出，UpTouch为向上敲入/敲出，DownTouch为向下敲入/敲出
            engine: 定价引擎，PricingEngine类，以下几种引擎均支持向上/向下、敲入/敲出、看涨/看跌8种单边障碍期权，支持敲出/未敲入现金返还
                            解析解和PDE引擎支持离散观察/连续观察，其余引擎只支持离散观察(默认为每日观察)
                    解析解: AnalyticBarrierEngine 敲入现金返还为到期支付；敲出现金返还为立即支付
                    蒙特卡洛: MCBarrierEngine  敲入现金返还为到期支付；敲出现金返还为到期支付 todo: 敲出现金返还立即支付
                    PDE: FdmBarrierEngine  敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                    积分法: QuadBarrierEngine  敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                    二叉树: BiTreeBarrierEngine 敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用解析解定价引擎 AnalyticBarrierEngine - Merton(1973), Reiner & Rubinstein(1991a)
                                支持 Broadie, Glasserman和Kou(1995)均匀离散观察调整
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.barrier = barrier
        self.rebate = rebate  # 绝对值，敲出补偿，或未敲入补偿
        self.parti = parti
        self._callput = callput
        self._inout = inout
        self._updown = updown
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)
        if payment_type is None:
            if self._inout == InOut.In:
                self.payment_type = PaymentType.Expire
            else:
                self.payment_type = PaymentType.Hit
        else:
            self.payment_type = payment_type
            if self._inout == InOut.In:
                assert payment_type == PaymentType.Expire, "ValueError: 敲入期权的现金返还方式PaymentType一定是到期支付Expire"
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = AnalyticBarrierEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if observable == self.engine.process.spot:
            if self.discrete_obs_interval is None:  # TODO: 目前update仅支持连续观察，离散观察需要判断是否是观察日
                if observable.value >= self.barrier and self.updown == UpDown.Up:
                    self.status = StatusType.UpTouch
                elif observable.value <= self.barrier and self.updown == UpDown.Down:
                    self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.discrete_obs_interval is None:
            obs_type = "连续观察"
        else:
            obs_type = "离散观察"
        return f"{obs_type}{self.barrier_type.value.text}期权"

    @property
    def callput(self):
        return self._callput

    @callput.setter
    def callput(self, value):
        self._callput = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)

    @property
    def updown(self):
        return self._updown

    @updown.setter
    def updown(self, value):
        self._updown = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)

    @property
    def inout(self):
        return self._inout

    @inout.setter
    def inout(self, value):
        self._inout = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入
        self.barrier_type = BarrierType.get_category(self._updown, self._inout, self._callput)

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍
            if self.updown == UpDown.Up:
                bound = (None, self.barrier)
            else:
                bound = (self.barrier, None)
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=bound,
                                                   rebate=(self.rebate, self.rebate))
        else:
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta
from contextlib import suppress
import numpy as np
from scipy.stats import norm, qmc
from ..processes import StochProcessBase
from ..utilities.enums import RandsMethod, LdMethod, ProcessType, EngineType
from ..utilities.patterns import Observer
from ..utilities.utility import logging
from .engine_base import PricingEngineBase


class McEngine(PricingEngineBase, Observer, metaclass=ABCMeta):
    """蒙特卡洛模拟定价引擎基类
    观察者，观察随机过程process对象，当process对象的属性变化时，自动更新状态，会重新生成价格路径"""
    engine_type = EngineType.McEngine

    def __init__(self, stoch_process: StochProcessBase = None, n_path=100000, rands_method=RandsMethod.LowDiscrepancy,
                 antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0, *,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            n_path: int，MC模拟路径数
            rands_method: 生成随机数方法，RandsMethod枚举类，Pseudorandom伪随机数/LowDiscrepancy低差异序列
            antithetic_variate: bool，是否使用对立变量法
            ld_method: 若使用了低差异序列，指定低差异序列方法，LdMethod枚举类，Sobol序列/Halton序列
            seed: int，随机数种子
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        # MC模拟路径数
        if n_path % 2 != 0 and antithetic_variate:
            n_path += 1
        self.n_path = n_path
        self.rands_method = rands_method  # 生成随机数方法
        self._ld_method = ld_method  # 低差异序列方法
        self._antithetic_variate = antithetic_variate  # 是否使用对立变量
        self.seed = seed  # 随机数种子
        self.__reset_rands = True  # 重置随机数标志位
        self.__reset_paths = True  # 重置路径标志位
        self.rands = None  # 随机数矩阵
        self.s_paths = None  # 价格路径矩阵
        self.var_paths = None  # 方差路径矩阵

    def set_stoch_process(self, stoch_process):
        """设置随机过程，先将自己从原来的随机过程的观察者列表中移除，再将自己加入新的随机过程的观察者列表"""
        with suppress(AttributeError, ValueError):
            self.process.remove_observer(self)
        self.process = stoch_process
        stoch_process.add_observer(self)  # 将本实例加入stoch_process观察者列表
        self.rands = None  # 更换随机过程之后需要重新生成随机数
        self.__reset_rands = True  # 更换随机过程之后需要重新生成随机数
        self.__reset_paths = True  # 更换随机过程之后需要重新生成路径

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.process.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """被观察者的值发生变化时，自动调用update方法，通知观察者"""
        self.__reset_paths = True  # process自动向观察者发送通知, 参数已改变, 重置路径标志位，重新生成路径

    def reset_paths_flag(self):
        """重置路径标志位，重新生成路径"""
        self.__reset_paths = True

    @property
    def antithetic_variate(self):
        return self._antithetic_variate

    @antithetic_variate.setter
    def antithetic_variate(self, new_value):
        self._antithetic_variate = new_value
        self.__reset_paths = True  # 重置路径标志位，重新生成路径
        self.__reset_rands = True  # 重置随机数标志位，重新生成随机数

    @property
    def ld_method(self):
        return self._ld_method

    @ld_method.setter
    def ld_method(self, new_value):
        self._ld_method = new_value
        self.__reset_paths = True  # 重置路径标志位，重新生成路径
        self.__reset_rands = True  # 重置随机数标志位，重新生成随机数

    def _randoms_generator(self, shape):  # 参数rands_method控制随机数获取方法，LD_method低差异序列生成方法；
        """生成随机数，返回shape形状的随机数矩阵
        如果需要的随机数矩阵的形状大于已有随机数矩阵的形状，则重新生成随机数矩阵，否则复用已有随机数矩阵
            rands_method: 控制随机数生成方法，包括伪随机数和低差异序列两种方法
            _ld_method: 控制低差异序列生成方法，包括Sobol和Halton两种低差异序列类型
        Args:
            shape: np.ndarray，(n_row, n_col)，随机数矩阵的形状
        Returns: self.rands，shape形状的随机数矩阵
        """
        if (self.rands is not None and (not self.__reset_rands)
                and self.rands.shape[0] >= shape[0] and self.rands.shape[1] >= shape[1]):
            logging.info("复用已有随机数")
            return self.rands[:shape[0], :shape[1]]
        else:
            np.random.seed(self.seed)
            if self.rands_method == RandsMethod.Pseudorandom:
                self.rands = np.random.standard_normal(shape)
            elif self.rands_method == RandsMethod.LowDiscrepancy:
                if self._ld_method == LdMethod.Sobol:
                    sampler = qmc.Sobol(d=shape[0], scramble=True, seed=self.seed)
                elif self._ld_method == LdMethod.Halton:
                    sampler = qmc.Halton(d=shape[0], scramble=True, seed=self.seed)
                else:
                    raise ValueError(
                        f'随机数生成方法输入错误，应为（Halton, Sobol）二者之一， 当前输入为{RandsMethod}')
                sampler.fast_forward(30)  # 跳过前三十项
                uniform_rands = sampler.random(shape[1]).transpose()
                np.random.shuffle(uniform_rands)
                self.rands = norm.ppf(uniform_rands)
            else:
                raise ValueError(
                    f'随机数生成方法输入错误，应为（RandsMethod.LowDiscrepancy, RandsMethod.Pseudorandom）二者之一， 当前输入为{RandsMethod}')
            self.__reset_rands = False  # 重置随机数标志位
            return self.rands

    def path_generator(self, n_step, spot=None, t_step_per_year=243):
        """根据标的种类和标的参数，生成价格路径，返回价格路径矩阵
        如果标的参数不变，且需要的价格路径矩阵的形状 ≤ 已有价格路径矩阵的形状，则复用已有价格路径矩阵，否则重新生成价格路径矩阵
        Args:
            n_step: int，价格路径的时间步数
            spot: float，标的期初价格
            t_step_per_year: int，每年的时间步数
        Returns: self.s_paths，价格路径矩阵
        """
        if (self.s_paths is not None and (not self.__reset_paths)
                and self.s_paths.shape[0] >= n_step + 1 and self.s_paths.shape[1] >= self.n_path
                and spot == self.process.spot()):
            logging.info("复用已有价格路径")
            return self.s_paths[:n_step + 1, :self.n_path]
        # else:
        return self._regenerate_paths(n_step, self.n_path, spot, t_step_per_year)

    def _regenerate_paths(self, n_step, n_path, spot, t_step_per_year):
        """生成价格路径的执行函数
        Args:
            n_step: int，价格路径的时间步数
            n_path: int，模拟路径数量
            spot: float，标的期初价格
            t_step_per_year: int，每年的时间步数
        Returns:
        """
        self.s_paths = np.empty(shape=(n_step + 1, n_path))
        dt = 1 / t_step_per_year

        if self.process() == ProcessType.BSProcess1D:
            if self.antithetic_variate:
                if n_path % 2 != 0:
                    n_path += 1
                half_n_path = n_path // 2
                self.rands = self._randoms_generator(shape=(n_step, half_n_path))
                rand_s = np.concatenate((self.rands, -self.rands), axis=1)
            else:
                self.rands = self._randoms_generator(shape=(n_step, n_path))
                rand_s = self.rands.copy()

            self.s_paths[0] = spot
            for step in range(1, n_step + 1):
                self.s_paths[step] = self.process.evolve(dt * step, self.s_paths[step - 1], dt, rand_s[step - 1])

        elif self.process() == ProcessType.Heston:
            self.var_paths = np.empty(shape=(n_step + 1, n_path))
            if self.antithetic_variate:
                self.rands = self._randoms_generator(shape=(n_step, n_path))
                rand_s = self.rands[:, :self.rands.shape[1] // 2].copy()
                rand_v = self.rands[:, self.rands.shape[1] // 2:].copy()
                rand_v = self.process.var_rho * rand_s + np.sqrt(1 - self.process.var_rho ** 2) * rand_v
                rand_s = np.concatenate((rand_s, -rand_s), axis=1)
                rand_v = np.concatenate((rand_v, -rand_v), axis=1)
            else:
                self.rands = self._randoms_generator(shape=(n_step, n_path * 2))
                rand_s = self.rands[:, :self.rands.shape[1] // 2].copy()
                rand_v = self.rands[:, self.rands.shape[1] // 2:].copy()
                rand_v = self.process.var_rho * rand_s + np.sqrt(1 - self.process.var_rho ** 2) * rand_v
            self.s_paths[0] = spot
            self.var_paths[0] = self.process.v0

            for step in range(1, n_step + 1):
                [self.s_paths[step], self.var_paths[step]] = self.process.evolve(dt * step, [self.s_paths[step - 1],
                                                                                             self.var_paths[step - 1]],
                                                                                 dt,
                                                                                 [rand_s[step - 1], rand_v[step - 1]])
            self.var_paths = np.maximum(self.var_paths, 0)  # heston方差非负修正：部分截断形式
        else:
            raise ValueError(f'随机过程类型输入错误，应为（ProcessType.BSProcess1D, ProcessType.Heston）二者之一，'
                             f'当前输入为{self.process()}')

        self.__reset_paths = False  # 重置路径标志位
        return self.s_paths



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from dateutil import rrule
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.utility import time_this
from pricelib.common.product_base import OptionBase
from pricelib.pricing_engines.analytic_engines.cashflow_engine import CashFlowEngine

__all__ = ['CashFlow']


class CashFlow(OptionBase):
    """固定现金流现值，相当于零息债券现值。"""

    def __init__(self, payment_date=None, cashflow=None, *, engine=None, trade_calendar=CN_CALENDAR,
                 annual_days=AnnualDays.N365, r=None):
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.payment_date = payment_date
        self.cashflow = cashflow
        if engine is not None:
            self.set_pricing_engine(engine)
        elif r is not None:
            default_engine = CashFlowEngine(s=cashflow, r=r)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        self.engine = engine

    def __repr__(self):
        """返回期权的描述"""
        return "固定现金流现值"

    @time_this
    def price(self, t: datetime.date = None):
        """计算现金流现值
        Args:
            t: datetime.date，计算期权价格的日期
        Returns: 现金流现值
        """
        calculate_date = global_evaluation_date() if t is None else t

        n_year = rrule.rrule(rrule.YEARLY, dtstart=calculate_date, until=self.payment_date).count()
        n_years_later = calculate_date.replace(year=calculate_date.year + n_year)
        n_natural_days_per_year = (n_years_later - calculate_date).days / n_year

        maturity = (self.payment_date - calculate_date).days / n_natural_days_per_year
        self.engine.get_product_params(maturity, self.cashflow)
        price = self.engine.calc_present_value(self)
        return price

    def delta(self, *args, **kwargs):
        delta = 0
        return delta

    def gamma(self, *args, **kwargs):
        gamma = 0
        return gamma

    def vega(self, *args, **kwargs):
        vega = 0
        return vega

    def theta(self, t: datetime.date = None, *args, **kwargs):
        calculate_date = global_evaluation_date() if t is None else t
        theta = self.price() * -self.engine.process.interest(calculate_date)
        return theta

    def rho(self, *args, **kwargs):
        rho = 0
        return rho



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticPortfolioEngine(AnalyticEngine):
    """香草组合解析解定价引擎
    该类没有实现具体的计算方法，只是为了提供engine_type = EngineType.AnEngine的引擎类型，
    这样在计算时，会调用各个香草期权的解析解定价引擎
    """

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from enum import Enum, unique
import numpy as np
from scipy.interpolate import interp1d
from ..utilities.patterns import Observable, SimpleQuote


class ConstantRate(SimpleQuote):
    """无风险利率/分红率常数类"""

    def disc_factor(self, t2, t1=0.):
        """返回指定时间t2对应的折现因子，或者t1到t2的折现因子(t1 < t2)
        Args:
            t1: 可以是float或numpy.ndarray, 默认为0。
            t2: 可以是float或numpy.ndarray, 但是t1和t2都是array时，维数必须一致
        Returns:
            float, 折现因子
        """
        return np.exp(-self(t2) * t2 + self(t1) * t1)


class RateTermStructure(Observable):
    """无风险利率/分红融券率曲线，线性插值"""

    def __init__(self, t0, fn, name=None):
        super().__init__()
        self.t0 = t0  # rate的初始时间
        self.__fn_rate = fn  # rate关于时间的函数
        self.__name = name

    @staticmethod
    def _interp_fn(tenors, rates):
        """返回曲线的线性插值函数
        Args:
            tenors: 期限数组
            rates: 无风险利率/分红融券率数组
        Returns:
            function，曲线的线性插值函数
        """
        assert len(tenors) == len(rates), "Error: 输入曲线的t与value数组长度不一致"
        linear_rate = interp1d(tenors, rates, kind='linear', bounds_error=False, fill_value=(rates[0], rates[-1]))
        return linear_rate

    def update_fn_rate(self, tenors, rates):
        """更新rate关于时间的函数，自动通知观察者"""
        self.__fn_rate = self._interp_fn(tenors, rates)
        self.notify_observers(name=self.__name)  # 自动向观察者发送通知

    @classmethod
    def from_array(cls, tenors, rates):
        """新建一个 RateTermStructure 对象 - 方法1: 用数组新建
        tenors是时间点数组, divs是分红率数组"""
        linear_rate = cls._interp_fn(tenors, rates)
        return cls(tenors[0], lambda t: linear_rate(t))

    @classmethod
    def from_fn(cls, t0, fn):
        """用函数新建一个 RateTermStructure 对象 - 方法2: 直接用函数新建
        t0是起始日，fn是rate关于时间的函数"""
        return cls(t0, fn)

    def __call__(self, t):
        """返回指定时间t对应的rate，t可以是float或numpy.arrays"""
        return self.__fn_rate(t)

    def disc_factor(self, t2, t1=0.):
        """返回指定时间t2对应的折现因子，或者t1到t2的折现因子
        t1可以是float或numpy.ndarray, 默认为0。
        t2可以是float或numpy.ndarray, 但是t1和t2都是array时，维数必须一致。"""
        return np.exp(-self(t2) * t2 + self(t1) * t1)

    @property
    def name(self):
        return self.__name


@unique
class InterpolationMode(Enum):
    LinearZero = 0  # 初始时刻t0是0，discs为折现到t0的折现因子，取对数除以时间后，对r线性插值，时间是距离起始时间t0的时长，两侧水平外推
    LinearLogDiscount = 1  # 折现因子为exp(-rt)，取对数，对-rt线性插值，时间是年化时间点，两侧线性外推


class DiscountCurve:
    """
    折现因子曲线。假设即期利率曲线分段线性，即每个标准期限即期利率线性相连。
    利率互换即期利率曲线为SC(t)，贴现函数曲线为DF(t)，满足DF(t)=exp^{-SC(t)*t}
    每期的贴现现金流应该等于初始的本金100，应有 100 = PV(k) = 利息部分 + 本金部分
        利息部分 = 100 * SC(k) * sum_{i=1}^k [T(i) - T(i-1)] * DF(i)
        本金部分 = 100 x DF(k)
        其中i为第i个付息期
    """

    def __init__(self, t0, fn):
        self.t0 = t0  # 初始时间
        self.__discount_factor = fn  # 折现因子插值函数

    @classmethod
    def from_array(cls, tenors, discs, interp_mode=InterpolationMode.LinearZero):
        """新建一个 DiscountCurve 对象 - 方法1: 用数组新建
        tenors是时间点数组, discs是折现因子数组, interp_mode是插值方式
            LinearZero 初始时刻t0是0，discs为折现到t0的折现因子，取对数除以时间后，对r线性插值，时间是距离起始时间t0的时长，两侧水平外推
            LinearLogDiscount 折现因子为exp(-rt)，取对数，对-rt线性插值，时间是年化时间点，两侧线性外推"""
        assert len(tenors) == len(discs), "Error: 输入曲线的t与value数组长度不一致"
        if interp_mode == InterpolationMode.LinearZero:
            y = -np.log(discs[1:]) / (tenors[1:] - tenors[0])
            linzero = interp1d(tenors[1:], y, kind='linear', bounds_error=False, fill_value=(y[0], y[-1]))
            return cls(tenors[0], lambda t: np.exp(-linzero(t) * t))
        if interp_mode == InterpolationMode.LinearLogDiscount:
            linlogdisc = interp1d(tenors, np.log(discs), kind='linear', bounds_error=False, fill_value='extrapolate')
            return cls(tenors[0], lambda t: np.exp(linlogdisc(t)))

        raise NotImplementedError(f"不支持的插值方式{interp_mode}")

    @classmethod
    def from_fn(cls, t0, fn):
        """用函数新建一个 DiscountCurve 对象 - 方法2: 直接用函数新建
        t0是起始日，fn是折现因子函数"""
        return cls(t0, fn)

    def __call__(self, t, dt=3e-3):
        """经过差分计算，泰勒展开e^x-1=x，得到的t处利率rt
        其中self(t)：t处的折现因子exp(-rt), dt=3e-3约等于1/365，即一个自然日
        """
        return (self(t - dt) / self(t + dt) - 1) / dt / 2

    def disc_factor(self, t2, t1=0.):
        """返回指定时间t2对应的折现因子，或者t1到t2的折现因子
        t1可以是float或numpy.ndarray, 默认为0。
        t2可以是float或numpy.ndarray, 但是t1和t2都是array时，维数必须一致。"""
        return self.__discount_factor(t2) / self.__discount_factor(t1)


if __name__ == "__main__":
    div = RateTermStructure.from_array([0, 1 / 12, 2 / 12, ], [0.02, 0.15, 0.18])
    r = RateTermStructure.from_array([0, 1 / 12, 2 / 12], [0.015, 0.02, 0.025])
    tau = 0.5
    print(r(tau) - div(tau))



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from scipy.stats import norm
from pricelib.common.processes import StochProcessBase
from pricelib.common.utilities.enums import ExerciseType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticVanillaEuEngine(AnalyticEngine):
    """欧式期权BSM解析解定价引擎"""

    def __init__(self, stoch_process: StochProcessBase = None, *, s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        # 以下属性指向需要定价的产品，由calc_present_value方法设置
        self.prod = None

    def d1(self, tau, spot):
        """BSM公式d1"""
        d1_value = (np.log(spot / self.prod.strike) + (self.process.drift(tau) +
                                                       0.5 * self.process.diffusion(tau, spot) ** 2) * tau) / (
                           self.process.diffusion(tau, spot) * np.sqrt(tau))
        return d1_value

    def d2(self, tau, spot):
        """BSM公式d2, d2 = d1 - sigma * sqrt(t)"""
        d2_value = self.d1(tau, spot) - self.process.diffusion(tau, spot) * np.sqrt(tau)
        return d2_value

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert prod.exercise_type == ExerciseType.European, "Error: 欧式期权解析解只支持European行权类型."
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()

        r = self.process.interest(tau)
        q = self.process.div(tau)
        price = (prod.callput.value * (spot * math.exp(-q * tau) * norm.cdf(prod.callput.value * self.d1(tau, spot))
                                       - prod.strike * math.exp(-r * tau) * norm.cdf(
                    prod.callput.value * self.d2(tau, spot))))
        return price

    def delta(self, prod, spot=None, *args, **kwargs):
        """∂V/∂S"""
        self.prod = prod
        tau = (prod.end_date - global_evaluation_date()).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        q = self.process.div(tau)
        delta = self.prod.callput.value * math.exp(-q * tau) * norm.cdf(self.prod.callput.value * self.d1(tau, spot))
        return delta

    def gamma(self, prod, spot=None, *args, **kwargs):
        """∂2V/∂S2"""
        self.prod = prod
        tau = (prod.end_date - global_evaluation_date()).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        gamma = math.exp(-q * tau) * norm.pdf(self.d1(tau, spot)) / (spot * vol * np.sqrt(tau))
        return gamma

    def vega(self, prod, spot=None, *args, **kwargs):
        """∂V/∂σ"""
        self.prod = prod
        tau = (prod.end_date - global_evaluation_date()).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        q = self.process.div(tau)
        vega = math.exp(-q * tau) * spot * norm.pdf(self.d1(tau, spot)) * np.sqrt(tau)
        return vega

    def rho(self, prod, spot=None, *args, **kwargs):
        """∂V/∂r"""
        self.prod = prod
        tau = (prod.end_date - global_evaluation_date()).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        rho = self.prod.callput.value * self.prod.strike * tau * np.exp(-r * tau) * norm.cdf(
            self.prod.callput.value * self.d2(tau, spot))
        return rho

    def theta(self, prod, spot=None, *args, **kwargs):
        """∂V/∂t"""
        self.prod = prod
        tau = (prod.end_date - global_evaluation_date()).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        sign = self.prod.callput.value
        theta = (-1 * spot * math.exp(-q * tau) * norm.pdf(self.d1(tau, spot)) * vol / (2 * np.sqrt(tau))
                 - sign * r * self.prod.strike * np.exp(-r * tau) * norm.cdf(sign * self.d2(tau, spot))
                 + sign * q * spot * np.exp(-q * tau) * norm.cdf(sign * self.d1(tau, spot)))
        return theta



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
import numpy as np
from pricelib.common.utilities.enums import CallPut
from pricelib.common.time import (AnnualDays, CN_CALENDAR)
from pricelib.common.utilities.utility import time_this, logging
from pricelib.pricing_engines.integral_engines import QuadAutoCallEngine
from .autocallable_base import AutocallableBase


class AutoCall(AutocallableBase):
    """Autocall Note(二元小雪球)，不带敲入的自动赎回结构，每月观察敲出，到期未敲出获得红利票息(保本产品)"""

    def __init__(self, s0, *, barrier_out, coupon_out, coupon_div=0, callput=CallPut.Call, lock_term=1,
                 maturity=None, start_date=None, end_date=None, obs_dates=None, pay_dates=None, margin_lvl=1,
                 engine=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            s0: float，标的初始价格
            barrier_out: float，敲出障碍价，绝对值/百分比
            coupon_out: float，敲出票息，百分比，年化
            coupon_div: float，红利票息，百分比，年化
            callput: 看涨看跌，CallPut枚举类，默认为Call
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            margin_lvl:  float，保证金比例，默认为1，即无杠杆
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCAutoCallEngine
                    PDE: FdmSnowBallEngine
                    积分法: QuadAutoCallEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，长度需要与敲出观察日数一致，如不指定则默认为敲出观察日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用 积分法 定价引擎 QuadAutoCallEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, annual_days=annual_days)
        len_obs_dates = len(self.obs_dates.date_schedule)
        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.coupon_out = np.ones(len_obs_dates) * coupon_out
        self.coupon_div = coupon_div
        self.barrier_in = np.zeros(len_obs_dates)  # 小雪球无敲入
        self.callput = callput
        self.margin_lvl = margin_lvl  # 预付金比例
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = QuadAutoCallEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        self.engine = engine
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def __repr__(self):
        """返回期权的描述"""
        return f"Auto{self.callput.name} Note(二元小雪球)"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta, abstractmethod


class StochProcessBase(metaclass=ABCMeta):
    """动态过程抽象类"""

    @abstractmethod
    def evolve(self, t, x, dt, dw):
        """随机微分方程离散化，递推一步dt的演化函数
        Args:
            t: float，当前时间
            x: np.ndarray，当前时间的价格向量
            dt: float，时间增量
            dw: np.ndarray，布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        Returns:
        """

    @abstractmethod
    def drift(self, *args, **kwargs):
        """随机微分方程漂移项dt系数"""

    @abstractmethod
    def diffusion(self, *args, **kwargs):
        """随机微分方程扩散项dW系数"""

    @abstractmethod
    def get_fn_pde_coef(self, *args, **kwargs):
        """返回PDE系数组件abc的函数，设pde有限差分的价格向量索引为i_vec，
            a是需要乘以i_vec^2的系数，b是需要乘以i_vec的系数，c是不需要乘以i_vec的系数"""



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.time import global_evaluation_date
from pricelib.common.utilities.enums import TouchType, ExerciseType, PaymentType
from pricelib.common.pricing_engine_base import McEngine


class MCDoubleDigitalEngine(McEngine):
    """双边二元期权(欧式-二元凸式/二元凹式；美式-双接触/双不接触) Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
         Args:
             prod: Product产品对象
             t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
             spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
         Returns: float，现值
         """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date)
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径
        r = self.process.interest(_maturity)
        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        # 美式：美式双接触/美式双不接触
        if prod.exercise_type == ExerciseType.American:
            if prod.discrete_obs_interval is None:  # 每日观察
                obs_points = np.arange(0, _maturity_business_days + 1, 1)[1:]
            else:  # 均匀离散观察
                dt_step = prod.discrete_obs_interval * prod.t_step_per_year
                obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
                obs_points = np.concatenate((np.array([0]), obs_points))
                obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

            if prod.touch_type == TouchType.NoTouch:  # 美式双不接触一定是到期支付
                strike_bool = np.where((paths[obs_points] <= prod.bound[0]) | (paths[obs_points] >= prod.bound[1]),
                                       0, 1)
                value = prod.rebate[0] * math.exp(-r * _maturity) * np.mean(np.min(strike_bool, axis=0))
            elif prod.touch_type == TouchType.Touch:
                step_index = np.tile(obs_points, (self.n_path, 1)).T
                hit_lower = np.min(
                    np.where(paths[obs_points] <= prod.bound[0], step_index.astype(int), _maturity_business_days + 1),
                    axis=0)
                hit_upper = np.min(
                    np.where(paths[obs_points] >= prod.bound[1], step_index.astype(int), _maturity_business_days + 1),
                    axis=0)
                if prod.payment_type == PaymentType.Hit:
                    value = np.mean(
                        np.where(hit_lower < hit_upper, prod.rebate[0] * np.exp(-r * hit_lower / prod.t_step_per_year),
                                 np.where(hit_lower > hit_upper,
                                          prod.rebate[1] * np.exp(-r * hit_upper / prod.t_step_per_year), 0)))
                elif prod.payment_type == PaymentType.Expire:
                    value = np.mean(np.where(hit_lower < hit_upper, prod.rebate[0],
                                             np.where(hit_lower > hit_upper, prod.rebate[1],
                                                      0))) * math.exp(-r * _maturity)
                else:
                    raise ValueError("PaymentType must be Hit or Expire")
            else:
                raise ValueError("TouchType must be Touch or NoTouch")
        # 欧式：二元凹式/二元凸式
        elif prod.exercise_type == ExerciseType.European:
            if prod.touch_type == TouchType.Touch:
                payoff = np.where(paths[-1] <= prod.bound[0], prod.rebate[0],
                                  np.where(paths[-1] >= prod.bound[1], prod.rebate[1], 0))
            elif prod.touch_type == TouchType.NoTouch:
                payoff = ((paths[-1] > prod.bound[0]) & (paths[-1] < prod.bound[1])) * prod.rebate[0]
            else:
                raise ValueError("TouchType must be Touch or NoTouch")
            value = np.mean(payoff) * math.exp(-r * _maturity)
        else:
            raise ValueError("ExerciseType must be American or European")
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import BiTreeEngine
from pricelib.common.time import global_evaluation_date


class BiTreePortfolioEngine(BiTreeEngine):
    """香草组合二叉树定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        n_step = int(tau * self.tree_branches)

        s_paths = self.path_generator(n_step, tau, spot).copy()
        v_grid = np.zeros(shape=s_paths.shape)
        for vanilla, position in prod.vanilla_list:
            v_grid[:, n_step] += np.maximum(vanilla.callput.value * (s_paths[:, -1] - vanilla.strike), 0) * position
        for j in range(n_step - 1, -1, -1):
            v_grid[:j + 1, j] = (v_grid[:j + 1, j + 1] * self.p + v_grid[1:j + 2, j + 1] * (
                    1 - self.p)) * np.exp(-self.process.interest(tau) * self.dt)
        return v_grid[0, 0]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.time import global_evaluation_date


class QuadFCNEngine(QuadEngine):
    """FCN/DCN 数值积分定价引擎"""

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date) / prod.t_step_per_year
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        obs_dates = np.array([num for num in obs_dates if num >= 0]) / prod.t_step_per_year
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        _barrier_out = prod.barrier_out[-len(obs_dates):].copy()
        _barrier_in = prod.barrier_in[-len(obs_dates):].copy()
        _barrier_yield = prod.barrier_yield[-len(obs_dates):].copy()
        _coupon = prod.coupon[-len(obs_dates):].copy()

        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(_maturity)
        q = self.process.div(_maturity)
        vol = self.process.vol(_maturity, spot)
        self._check_method_params()
        self.set_quad_params(r=r, q=q, vol=vol)
        if obs_dates[0] == 0:
            # 如果估值日就是敲出观察日
            if spot >= _barrier_out[0]:  # 发生敲出
                return (prod.margin_lvl + _coupon[0]) * prod.s0
            else:  # 没有发生敲出，则积分迭代步数少一次
                s0_dt = obs_dates[1]
                dt_vec = np.diff(obs_dates)[1:]
        else:
            s0_dt = obs_dates[0]
            dt_vec = np.diff(obs_dates)

        backward_steps = dt_vec.size
        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, _maturity_business_days)
        s_vec = np.exp(self.ln_s_vec)

        v_grid = np.zeros(shape=(self.n_points, backward_steps + 2))

        barrier_out_idx = np.searchsorted(s_vec, _barrier_out, side='right')
        barrier_yield_idx = np.searchsorted(s_vec, _barrier_yield, side='right')

        for step in range(backward_steps + 1, 0, -1):
            if step == backward_steps + 1:  # 设置期末价值状态: 预付金 + 派息 + 到期敲入或有亏损
                v_grid[:, -1] = (prod.margin_lvl * prod.s0 +
                                 np.where(s_vec >= _barrier_yield[-1], _coupon[-1] * prod.s0, 0) +
                                 np.where(s_vec >= _barrier_in[-1], 0, prod.parti_in * (- prod.strike_upper +
                                          np.where(s_vec > prod.strike_lower, s_vec, prod.strike_lower))))
            else:
                # 数值积分，计算前一个敲出观察日的期权价值
                v_grid[:, step] = self.fft_step_backward(self.ln_s_vec, self.ln_s_vec,
                                                         v_grid[:, step + 1], dt_vec[step - 1])
                # 敲出价之上，发生敲出，返还预付金 + 派息
                v_grid[barrier_out_idx[step - 1]:, step] = (prod.margin_lvl + _coupon[step - 1]) * prod.s0
                # 派息线之上，敲出价之下，派息
                v_grid[barrier_yield_idx[step - 1]:barrier_out_idx[step - 1], step] += _coupon[step - 1] * prod.s0
                # 派息线之下，不派息

        x = np.array([np.log(spot)])
        value = self.fft_step_backward(x, self.ln_s_vec, v_grid[:, 1], s0_dt)[0]
        if obs_dates[0] == 0:  # 如果估值日就是敲出观察日，并且spot位于派息线上、敲出线下，则需要加上派息的价值
            value += _coupon[0] * prod.s0
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import InOut, PaymentType
from .fdm_barrier_engine import FdmBarrierEngine


class FdmDoubleSharkEngine(FdmBarrierEngine):
    """双鲨结构PDE有限差分法定价引擎"""
    inout = InOut.Out  # 双鲨结构属于敲出期权

    def _init_boundary_condition(self, smax, maturity):
        """初始化边界条件
        双鲨结构，上方边界条件是call，下方边界条件是put
        Args:
            smax: float, 有限差分网格上边界的价格最大值
            maturity: float, 到期时间
        Returns:
            self.fn_callput: List[Callable], 0~smax网格的边界条件
            self.fn_bound: List[Callable], 障碍价格网格的边界条件
        """
        # 生成无障碍价格的边界条件：在s=0处，call近似为0; 在smax处，put近似为0。根据看涨看跌平价 call-put = S*exp(-qT) - K*exp(-rT),
        # 可得对应的put(0) = K * exp(-rT), call(smax) = smax * exp(-qT) - K * exp(-rT)
        self.fn_callput = [
            lambda u: (self.prod.strike[0] * self.process.interest.disc_factor(maturity, u)) * self.prod.parti[0],
            lambda u: (smax * self.process.div.disc_factor(maturity, u) -
                       self.prod.strike[1] * self.process.interest.disc_factor(maturity, u)) * self.prod.parti[1]]
        # 生成障碍价格的Dirichlet边界条件： 第u天触碰障碍产生的补偿，折现到第u天的价值。
        # 返回一个包含两个lambda函数的列表，列表索引分别为[0, 1]，每个lambda函数的形参r=self.rebate中的r补偿值
        # 需要注意的是，敲出期权rebate[0]是下边界补偿，敲出期权rebate[1]是上边界补偿；敲入期权rebate[0]是敲入补偿，敲入期权rebate[1]没有用处。
        self.fn_bound = []
        for r in self.rebate:
            if self.prod.payment_type == PaymentType.Hit:  # 触碰障碍立即支付
                self.fn_bound.append(lambda u: r)
            else:  # 'PaymentMethod.Expire' 触碰障碍后，到期时支付
                self.fn_bound.append(lambda u: r * self.process.interest.disc_factor(maturity, u))

    def _init_terminal_condition(self, s_vec):
        """初始化终止条件
        双鲨结构，上方是call，下方是put"""
        call_put_payoff = (np.maximum(s_vec - self.prod.strike[1], 0) * self.prod.parti[1] +
              np.maximum(self.prod.strike[0] - s_vec, 0) * self.prod.parti[0])
        yv = np.where(s_vec >= self.bound[1], self.rebate[1],
                      np.where(s_vec <= self.bound[0], self.rebate[0], call_put_payoff))
        return yv



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .vanilla_option import *
from .underlying import *
from .vanilla_portfolio import *

__all__ = ['VanillaOption', 'VanillaPortfolio', 'UnderlyingAsset', 'Spread', 'Butterfly', 'Straddle', 'Strangle',
           'Condor', 'DiscountCall', 'RiskReversal', 'Collar']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class StandardSnowball(AutocallableBase):
    """经典雪球产品类"""

    def __init__(self, s0, barrier_out, barrier_in, coupon_out, obs_dates=None, pay_dates=None, coupon_div=None,
                 lock_term=3, parti_in=1, margin_lvl=1, status=StatusType.NoTouch, engine=None, maturity=None,
                 start_date=None, end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365,
                 t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            s0: float，标的初始价格
            barrier_out: float，敲出障碍价，绝对值/百分比
            barrier_in: float，敲入障碍价，绝对值/百分比
            coupon_out: float，敲出票息，百分比，年化
            coupon_div: float，红利票息，百分比，年化
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            status: 敲入敲出状态，StatusType枚举类，默认为NoTouch未敲入未敲出，UpTouch为已敲出，DownTouch为已敲入
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCAutoCallableEngine
                    PDE: FdmSnowBallEngine
                    积分法: QuadSnowballEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，长度需要与敲出观察日数一致，如不指定则默认为敲出观察日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用PDE定价引擎 FdmSnowBallEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, parti_in=parti_in,
                         margin_lvl=margin_lvl, engine=engine, s=s, r=r, q=q, vol=vol)
        len_obs_dates = len(self.obs_dates.date_schedule)
        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in
        self.coupon_out = np.ones(len_obs_dates) * coupon_out
        self.coupon_div = coupon_div if coupon_div is not None else self.coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = 0

    def __repr__(self):
        """返回期权的描述"""
        return "经典雪球(平敲雪球)"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import RandsMethod, LdMethod, StatusType
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCAutoCallableEngine(McEngine):
    """自动赎回结构(雪球类) Monte Carlo 模拟定价引擎
    支持变敲出、变敲入、变票息等要素可变型雪球结构"""

    def __init__(self, stoch_process=None, n_path=100000, rands_method=RandsMethod.LowDiscrepancy,
                 antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0, *,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            n_path: int，MC模拟路径数
            rands_method: 生成随机数方法，RandsMethod枚举类，Pseudorandom伪随机数/LowDiscrepancy低差异序列
            antithetic_variate: bool，是否使用对立变量法
            ld_method: 若使用了低差异序列，指定低差异序列方法，LdMethod枚举类，Sobol序列/Halton序列
            seed: int，随机数种子
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, n_path, rands_method=rands_method, antithetic_variate=antithetic_variate,
                         ld_method=ld_method, seed=seed, s=s, r=r, q=q, vol=vol)
        # 以下为计算过程的中间变量
        self.prod = None  # Product产品对象
        self.obs_dates = None  # 根据估值日，将敲出观察日转化为List[int]，交易日期限
        self.pay_dates = None  # 根据起始日，将支付日转化为List[float]，年化自然日期限，用于计算票息
        self.pay_dates_tau = None  # 根据估值日，将支付日转化为List[float]，年化自然日期限，用于折现
        self._maturity = None  # 年化到期时间
        self.knock_out_date = None  # 每条路径的敲出时间
        self.not_knock_out = None  # 每条路径是否未敲出
        self.knock_in_scenario = None  # 每天路径是否敲入
        self.knock_out_profit = None  # 敲出payoff加总
        self.hold_to_maturity_profit = None  # 持有到期payoff加总
        self.knock_in_loss = None  # 敲入payoff加总
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        self._barrier_out = None
        self._barrier_in = None
        self._coupon_out = None

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        self._maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        self.obs_dates = np.array([num for num in obs_dates if num >= 0])
        # 支付日 - 计算起始日到支付日的天数，用于计算应支付的敲出收益
        calculate_start_diff = (calculate_date - prod.start_date).days
        pay_dates = prod.pay_dates.count_calendar_days(prod.start_date)
        self.pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= calculate_start_diff])
        assert len(self.obs_dates) == len(self.pay_dates), f"Error: {prod}的观察日和付息日长度不一致"
        # 支付日 - 计算估值日到支付日的天数，用于折现
        pay_dates_tau = prod.pay_dates.count_calendar_days(calculate_date)
        self.pay_dates_tau = np.array([num / prod.annual_days.value for num in pay_dates_tau if num >= 0])
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        self._barrier_out = prod.barrier_out[-len(self.obs_dates):].copy()
        self._barrier_in = prod.barrier_in[-len(self.obs_dates):].copy()
        self._coupon_out = prod.coupon_out[-len(self.obs_dates):].copy()
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        self.s_paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                           t_step_per_year=prod.t_step_per_year).copy()
        # 统计各个情景的payoff
        self._cal_knock_out_date()
        self._cal_knock_out_payoff()
        self._cal_knock_in_scenario(calculate_date)
        self._cal_hold_to_maturity_payoff()
        self._cal_knock_in_payoff()
        result = (self.knock_out_profit + self.hold_to_maturity_profit + self.knock_in_loss) / self.n_path
        return result

    def _cal_knock_out_date(self):
        """计算每条路径敲出时间"""
        # 敲出部分
        knock_out_scenario = np.tile(self.obs_dates, (self.n_path, 1)).T
        barrier_out = np.array(self._barrier_out)
        barrier_out = np.tile(barrier_out, (self.n_path, 1)).T
        # 记录每条路径的具体敲出日，如果无敲出则保留inf      注意：路径矩阵包含了期初S0的行
        knock_out_scenario = np.where(self.s_paths[self.obs_dates] >= barrier_out, knock_out_scenario.astype(int),
                                      np.inf)
        self.knock_out_date = np.min(knock_out_scenario, axis=0)
        self.not_knock_out = self.knock_out_date == np.inf

    def _cal_knock_out_payoff(self):
        """计算敲出部分的payoff"""
        prod = self.prod
        is_knock_out = self.knock_out_date != np.inf
        # 每个交易日敲出票息
        coupon_call_array = np.array(self._coupon_out).repeat(
            np.diff(np.append(np.zeros((1,)), self.obs_dates)).astype(int))
        # 每条路径敲出交易日序列
        knock_out_time = self.knock_out_date[is_knock_out].astype(int)
        coupon_call = coupon_call_array[knock_out_time - 1]
        # 每个交易日对应的自然日的下一个敲出观察自然日
        next_calendar_day = np.array(self.pay_dates).repeat(
            np.diff(np.append(np.zeros((1,)), self.obs_dates)).astype(int))
        next_calendar_day = np.append(self.pay_dates[0], next_calendar_day)
        # 所有敲出路径的起息日到敲出支付日的时间，年化，计算支付票息用
        knock_out_time_annual = next_calendar_day[knock_out_time]
        # 所有敲出路径的估值日到敲出支付日的时间，年化，折现用
        next_calendar_day_tau = np.array(self.pay_dates_tau).repeat(
            np.diff(np.append(np.zeros((1,)), self.obs_dates)).astype(int))
        next_calendar_day_tau = np.append(self.pay_dates_tau[0], next_calendar_day_tau)
        knock_out_tau = next_calendar_day_tau[knock_out_time]
        # 计算敲出价格，用于上涨参与payoff
        if isinstance(prod.strike_call, (int, float)):
            strike_call = prod.strike_call
        elif isinstance(prod.strike_call, (list, np.ndarray)) and len(prod.strike_call) == len(self.obs_dates):
            strike_call_array = (np.array(prod.strike_call)).repeat(
                np.diff(np.append(np.zeros((1,)), self.obs_dates)).astype(int))
            strike_call = strike_call_array[knock_out_time - 1]
        else:
            raise ValueError("敲出看涨执行价设置错误")
        # 把payoff先折现再求和,计算敲出总所入
        value = np.sum((prod.s0 * (coupon_call * (knock_out_time_annual if not prod.trigger else 1) + prod.margin_lvl)
                        + prod.parti_out * (self.s_paths[knock_out_time, is_knock_out] - strike_call))
                       * self.process.interest.disc_factor(knock_out_tau))
        self.knock_out_profit = value

    def _cal_knock_in_scenario(self, *args, **kwargs):
        """统计敲入的路径"""
        prod = self.prod
        if prod.status == StatusType.DownTouch:
            pass
        else:
            # 判断某一条路径是否有敲入
            if isinstance(prod.barrier_in, (int, float)):
                knock_in_level = prod.barrier_in
            elif isinstance(prod.barrier_in, (list, np.ndarray)):
                knock_in_level = np.array(self._barrier_in).repeat(
                    np.diff(np.append(np.zeros((1,)), self.obs_dates)).astype(int))
            else:
                raise ValueError("敲入线设置错误")
            knock_in_level = np.tile(knock_in_level, (self.n_path, 1)).T
            self.knock_in_scenario = np.any(self.s_paths[1:] <= knock_in_level, axis=0)

    def _cal_hold_to_maturity_payoff(self):
        """计算持有到期的payoff"""
        # 到期红利部分
        prod = self.prod
        if prod.status == StatusType.DownTouch:  # 已敲入
            self.hold_to_maturity_profit = 0
        else:
            # 持有到期，没有敲入也没有敲出
            hold_to_maturity = (~self.knock_in_scenario) & self.not_knock_out
            # 平稳持有到期路径条数
            hold_to_maturity_count = np.count_nonzero(hold_to_maturity)
            # 未敲出未敲入到期收入
            value = (hold_to_maturity_count * (prod.coupon_div * (self.pay_dates[-1] if not prod.trigger else 1)
                                               + prod.margin_lvl) * prod.s0 * self.process.interest.disc_factor(
                self._maturity))
            self.hold_to_maturity_profit = value

    def _cal_knock_in_payoff(self):
        """计算敲入部分的payoff"""
        prod = self.prod
        if prod.status == StatusType.DownTouch:  # 已敲入
            value = np.sum(np.maximum(np.minimum(self.s_paths[-1, self.not_knock_out] - prod.strike_upper, 0),
                                      prod.strike_lower - prod.strike_upper) * prod.parti_in
                           + prod.margin_lvl * prod.s0) * self.process.interest.disc_factor(self._maturity)
        else:
            value = np.sum(np.maximum(np.minimum(self.s_paths[-1, self.not_knock_out & self.knock_in_scenario]
                                                 - prod.strike_upper, 0),
                                      prod.strike_lower - prod.strike_upper) * prod.parti_in
                           + prod.margin_lvl * prod.s0) * self.process.interest.disc_factor(self._maturity)
        self.knock_in_loss = value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .analytic_engines import *
from .mc_engines import *
from .fdm_engines import *
from .tree_engines import *
from .integral_engines import *



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .barrier_option import *
from .double_barrier_option import *
from .double_shark import *
from .airbag import *

__all__ = ['BarrierOption', 'DoubleBarrierOption', 'DoubleShark', 'Airbag', ]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .bitree_vanilla_engine import *
from .bitree_barrier_engine import *
from .bitree_digital_engine import *
from .bitree_asian_engine import *
from .bitree_portfolio_engine import *

__all__ = ['BiTreeVanillaEngine', 'BiTreeBarrierEngine', 'BiTreeDigitalEngine', 'BiTreeAsianEngine',
           'BiTreePortfolioEngine', ]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta
from ..utilities.enums import EngineType
from .engine_base import PricingEngineBase


class AnalyticEngine(PricingEngineBase, metaclass=ABCMeta):
    """解析定价引擎基类"""
    engine_type = EngineType.AnEngine



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import TouchType, PaymentType, ExerciseType
from pricelib.common.pricing_engine_base import FdmEngine, FdmGridwithBound, FdmGrid
from pricelib.common.time import global_evaluation_date


class FdmDigitalEngine(FdmEngine):
    """二元(数字)期权PDE有限差分法定价引擎，整个存续期观察
    支持结构:
        单边二元: 欧式二元/美式二元/一触即付
        双边二元:
            欧式 - 二元凹式/二元凸式
            美式 - 美式双接触/美式双不接触
    """

    def __init__(self, stoch_process=None, s_step=800, n_smax=2, fdm_theta=1, *,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: StochProcessBase 随机过程
            s_step: int，价格步数
            n_smax: int，价格网格上界，设为n_smax倍初始价格
            fdm_theta: float，时间方向有限差分theta，0：explicit, 1: implicit, 0.5: Crank-Nicolson
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, s_step=s_step, n_smax=n_smax, fdm_theta=fdm_theta, s=s, r=r, q=q, vol=vol)
        # 以下为计算过程的中间变量
        self.prod = None  # Product产品对象
        self.fdm = None  # 有限差分网格对象，若连续观察，则为FdmGridwithBound；若离散观察则为FdmGrid
        self.bound = None  # 低/高障碍价格绝对数值
        self.rebate = None  # 低/高障碍价格对应补偿绝对数值
        self.fn_bound = None  # 边界条件函数

    def _gen_func_bound(self, bound_flag, rebate, maturity):
        """生成边界条件函数
        Args:
            bound_flag: 0或1，0代表边界上的期权价值是0，1代表边界上的期权价值是rebate
            rebate: float，二元期权的行权收益
            maturity: float，到期时间
        Returns:
            func_bound: function，边界条件函数，输入为年化时间，返回为期权价值
        """
        if self.prod.payment_type == PaymentType.Hit:  # 触碰障碍立即支付
            def func_bound(u):
                return bound_flag * rebate
        else:  # 'PaymentMethod.Expire' 触碰障碍后，到期时支付
            def func_bound(u):
                return bound_flag * rebate * self.process.interest.disc_factor(maturity, u)
        return func_bound

    def _init_boundary_condition(self, maturity):
        """
        生成障碍价格的Dirichlet边界条件： 第u天触碰障碍产生的payoff，折现到第u天的价值。
        返回一个包含两个lambda函数的列表，列表索引分别为[0, 1]，对应低障碍价和高障碍价。
        Args:
            maturity: float, 到期时间
        Returns: None
        """
        self.fn_bound = []
        flag_list = [1, 1]
        for i, rebate in enumerate(self.rebate):
            if self.bound[i] in [-np.inf, np.inf]:  # 障碍价无穷，即障碍价没有设置
                flag_list[i] *= -1  # 反转
            if self.prod.touch_type == TouchType.NoTouch:
                flag_list[i] *= -1  # 反转
            bound_flag = (1 + flag_list[i]) * 0.5  # 将±1转化为0或1
            self.fn_bound.append(self._gen_func_bound(bound_flag, rebate, maturity))

    def _init_terminal_condition(self, s_vec):
        """初始化终止条件
        Args:
            s_vec: List[float], 标的价格向量
        Returns:
            yv: List[float], 终止条件 - 期权价值向量
        """
        # 美式, 且连续观察
        if self.prod.exercise_type == ExerciseType.American and self.prod.discrete_obs_interval is None:
            # 连续观察，网格边界是障碍价，留到最后一列的都是一直没碰到边界的payoff
            if self.prod.touch_type == TouchType.Touch:
                yv = s_vec * 0
            else:  # NoTouch
                yv = s_vec * 0 + self.rebate[0]
            return yv
        if self.prod.exercise_type == ExerciseType.European or (  # 欧式或离散观察美式
                self.prod.exercise_type == ExerciseType.American and self.prod.discrete_obs_interval is not None):
            lower = s_vec <= self.bound[0]  # 小于低障碍价的标的价格bool索引
            upper = s_vec >= self.bound[1]  # 大于高障碍价的标的价格bool索引
            inner = (~lower) & (~upper)  # 大于低障碍价且小于于高障碍价，即在两个障碍价格之间的bool索引
            if self.prod.touch_type == TouchType.Touch:
                yv = s_vec * 0
                yv[lower] = self.rebate[0]
                yv[upper] = self.rebate[1]
            else:  # NoTouch
                yv = s_vec * 0
                yv[inner] = self.rebate[0]
            return yv

        raise ValueError(f"{self.prod.exercise_type}必须是欧式或美式")

    def calc_present_value(self, prod, t=None, spot=None, bound=(None, None), rebate=(0, 0)):
        """计算现值，PDE有限差分定价, 时间逆向迭代
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
            bound: (lower_barrier, upper_barrier), 低障碍价/高障碍价，障碍价格绝对数值。如果是默认值None，即为无穷大或无穷小inf。
            rebate: (lower_rebate, upper_rebate), 对敲出期权，元组第一位是低障碍价下方的补偿，第二位是高障碍价上方的补偿，
                                                        对敲入期权，元组第一位是敲入补偿，第二位没有用处。
        Returns: float，现值
        """
        self.prod = prod
        self.bound = np.array([(2 * i - 1) * np.inf if b is None else b for i, b in enumerate(bound)])  # 障碍价格区间
        self.rebate = np.array(rebate)  # 补偿回报

        calculate_date = global_evaluation_date() if t is None else t
        # 剩余到期时间，交易日
        _maturity = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()
        self.smax = self.n_smax * spot  # 价格网格上界, n_smax倍初始价格
        # 返回PDE系数组件abc的函数
        fn_pde_coef = self.process.get_fn_pde_coef(_maturity, spot)

        if prod.exercise_type == ExerciseType.American:
            if prod.discrete_obs_interval is None:  # 连续观察美式
                self.fdm = FdmGridwithBound(smax=self.smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                            fn_pde_coef=fn_pde_coef, bound=self.bound, fdm_theta=self.fdm_theta)
                self._init_boundary_condition(_maturity)
                self.fdm.set_boundary_condition(self.fn_bound)

                # 初始化到期日payoff价值
                yv = self._init_terminal_condition(self.fdm.s_vec)
                # 从后向前逆推
                yv = self.fdm.evolve_with_interval(_maturity, 0, yv)

            else:  # 离散观察美式
                self.fdm = FdmGrid(smax=self.smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                   fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta, maturity=_maturity)
                self._init_boundary_condition(_maturity)
                self.fdm.v_grid[0, :] = self.fn_bound[0](self.fdm.tv)
                self.fdm.v_grid[-1, :] = self.fn_bound[1](self.fdm.tv)

                lower = self.fdm.s_vec <= self.bound[0]  # 小于低障碍价的标的价格bool索引
                upper = self.fdm.s_vec >= self.bound[1]  # 大于高障碍价的标的价格bool索引
                t_step = round(prod.t_step_per_year * _maturity)  # 时间步数
                dt = 1 / prod.t_step_per_year
                obs_points = np.flip(np.round(prod.t_step_per_year *
                                              np.arange(_maturity, 0, -prod.discrete_obs_interval)).astype(int))
                obs_points = np.concatenate((np.array([0]), obs_points))

                yv = self._init_terminal_condition(self.fdm.s_vec)
                self.fdm.v_grid[1:-1, -1] = yv

                for j in range(t_step - 1, -1, -1):
                    yv = self.fdm.evolve(j, yv, dt)  # 每组(p, q)更新
                    if j in obs_points:
                        if prod.touch_type == TouchType.Touch:
                            if prod.payment_type == PaymentType.Expire:  # 到期支付
                                yv[lower] = self.rebate[0] * self.process.interest.disc_factor(_maturity,
                                                                                            j / t_step * _maturity)
                                yv[upper] = self.rebate[1] * self.process.interest.disc_factor(_maturity,
                                                                                            j / t_step * _maturity)
                            else:  # 立即支付
                                yv[lower] = self.rebate[0]
                                yv[upper] = self.rebate[1]
                        else:
                            yv[lower] = 0
                            yv[upper] = 0
                    self.fdm.v_grid[1:-1, j] = yv
        elif prod.exercise_type == ExerciseType.European:  # 欧式
            self.fdm = FdmGrid(smax=self.smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                               fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta, maturity=_maturity)
            self._init_boundary_condition(_maturity)
            self.fdm.v_grid[0, :] = self.fn_bound[0](self.fdm.tv)
            self.fdm.v_grid[-1, :] = self.fn_bound[1](self.fdm.tv)

            t_step = round(prod.t_step_per_year * _maturity)  # 时间步数
            dt = 1 / prod.t_step_per_year
            yv = self._init_terminal_condition(self.fdm.s_vec)
            self.fdm.v_grid[1:-1, -1] = yv

            for j in range(t_step - 1, -1, -1):
                yv = self.fdm.evolve(j, yv, dt)  # 每组(p, q)更新
                self.fdm.v_grid[1:-1, j] = yv
        else:
            raise ValueError(f"{prod.exercise_type}必须是欧式或美式")
        return np.float64(self.fdm.functionize(yv, kind="cubic")(spot))



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from pricelib.common.product_base.option_base import OptionBase
from ...common.utilities.enums import CallPut, AverageMethod, AsianAveSubstitution
from ...common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from ...common.utilities import time_this, logging
from pricelib.pricing_engines.analytic_engines import AnalyticAsianEngine
from pricelib.pricing_engines.mc_engines import MCAsianEngine


class AsianOption(OptionBase):
    """
    亚式期权：几何平均/算术平均；替代标的资产价格/替代执行价；增强亚式
        解析解只支持平均结算价，无增强亚式
        蒙特卡洛支持几何平均/算术平均，支持增强亚式，支持替代标的资产价格/替代执行价
        二叉树支持几何平均/算术平均，支持增强亚式，支持替代标的资产价格，不支持替代执行价
        todo:若定价时已经处于均价观察期内，需要获得当前已经实现的均价，并对行权价进行调整
    """

    def __init__(self, callput: CallPut, ave_method: AverageMethod, strike=None,
                 substitute=AsianAveSubstitution.Underlying, parti=1, enhanced=False, limited_price=None,
                 engine=None, maturity=None, start_date=None, end_date=None, obs_start=None, obs_end=None,
                 s_average=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """亚式期权初始化
        产品参数:
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            ave_method: 平均价计算方法，AverageMethod枚举类，几何平均Geometric/算术平均Arithmetic
            strike: float, 执行价
            substitute: 以均价代替标的资产价格/替代行权价，AsianAveSubstitution枚举类，Underlying/Strike
            parti: float, 参与率，默认为1
            enhanced: bool，是否为增强亚式期权，默认为False
            limited_price: float，增强型亚式限定标的资产价格的上/下限
            s_average: float，已实现的资产价格平均值(已经处于资产价格平均期内时)，默认为None
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticAsianEngine
                           几何平均替代结算价亚式: Kemma Vorst (1990) 连续观察，平均价观察期为整个期权有效期内
                           算术平均替代结算价亚式: Turnbull Wakeman (1991) 近似估计，平均观察期的起点任意，终点为期权到期日
                    二叉树: BiTreeAsianEngine 只支持算术平均替代结算价亚式
                    蒙特卡洛: MCAsianEngine
                           几何平均/算术平均替代结算价亚式
                           几何平均/算术平均替代执行价亚式
                           增强亚式
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_start: datetime.date，观察期起始日
            obs_end: datetime.date，观察期终止日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则平均替代收盘价亚式默认使用解析解定价引擎；平均替代执行价亚式和增强亚式默认使用蒙特卡洛定价引擎
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.obs_start = obs_start if obs_start is not None else self.start_date
        self.obs_end = obs_end if obs_end is not None else self.end_date
        self.parti = parti
        self.enhanced = enhanced
        self.limited_price = limited_price
        if enhanced:
            assert limited_price is not None, "Error: 增强型亚式必须输入限定标的价格上/下限参数"
        assert strike is not None or substitute == AsianAveSubstitution.Strike, "Error: 替代平均结算价，必需输入执行价参数"
        self.strike = strike
        self.callput = callput
        self.ave_method = ave_method
        self.substitute = substitute
        self.s_average = s_average
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            if substitute == AsianAveSubstitution.Underlying:
                default_engine = AnalyticAsianEngine(s=s, r=r, q=q, vol=vol)
                self.set_pricing_engine(default_engine)
            else:  # 其他情形：替代执行价亚式/增强亚式
                default_engine = MCAsianEngine(s=s, r=r, q=q, vol=vol)
                self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        self.engine = engine
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def __repr__(self):
        """返回期权的描述"""
        enhance_str = "增强型" if self.enhanced else ""
        return f"{self.ave_method.value}-{self.substitute.value}-{self.callput.name}{enhance_str}亚式期权"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot) * self.parti
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .mc_vanilla_engine import *
from .mc_barrier_engine import *
from .mc_digital_engine import *
from .mc_asian_engine import *
from .mc_portfolio_engine import *
from .mc_double_digital_engine import *
from .mc_double_barrier_engine import *
from .mc_double_shark_engine import *
from .mc_airbag_engine import *
from .mc_autocall_engine import *
from .mc_autocallable_engine import *
from .mc_phoenix_engine import *
from .mc_accumulator import *
from .mc_range_accural import *
from .mc_paris_snowball_engine import *

__all__ = ['MCVanillaEngine', 'MCBarrierEngine', 'MCDigitalEngine', 'MCDoubleDigitalEngine', 'MCAsianEngine',
           'MCPortfolioEngine', 'MCDoubleBarrierEngine', 'MCDoubleSharkEngine', 'MCAirbagEngine', 'MCAutoCallEngine',
           'MCAutoCallableEngine', 'MCPhoenixEngine', 'MCAccumulatorEngine', 'MCRangeAccuralEngine',
           'MCParisSnowballEngine']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCAirbagEngine(McEngine):
    """安全气囊 Monte Carlo 模拟定价引擎
    安全气囊是美式观察向下敲入期权，到期支付
    只支持离散观察(默认为每日观察)
    """

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径
        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        if prod.discrete_obs_interval is None:  # 每日观察
            obs_points = np.arange(0, _maturity_business_days + 1, 1)[1:]
        else:  # 均匀离散观察
            dt_step = prod.discrete_obs_interval * prod.t_step_per_year
            obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
            obs_points = np.concatenate((np.array([0]), obs_points))
            obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

        # 期末价值
        payoff = paths[-1] - prod.strike

        # 安全气囊，未敲入是call，已敲入是标的资产（上涨下跌可以设置不同的参与率）
        knock_in_bool = np.any(paths[obs_points] <= prod.barrier, axis=0)
        value = np.where(knock_in_bool,  # 敲入
                         np.where(payoff < 0, payoff * prod.knockin_parti,  # 敲入下跌参与率
                                  payoff * prod.reset_call_parti),  # 敲入重置后看涨参与率
                         np.where(payoff < 0, 0, payoff * prod.call_parti))  # 看涨参与率
        value = np.mean(value) * self.process.interest.disc_factor(_maturity)

        return value
